// VinciaShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Antenna class 
  
#ifndef VINCIA_VinciaAntenna_H
#define VINCIA_VinciaAntenna_H

#include "Basics.h"
#include "Event.h"
#include "PythiaStdlib.h"
#include "VinciaCommon.h"

namespace Vincia {
  
  // Classes contained in this header
  class Antenna;
  class AntennaSet;
  class SixInt;
  
  /*
   *  SixInt.h
   *  
   *
   *  Created by Mathias Ritzmann on 7/10/10.
   *  Pure Stubornness to use map with six integer indices
   *
   */
  
  class SixInt {
  public:
    // Constructor
  SixInt(int i1, int i2, int i3, int i4, int i5, int i6) :
    int1(i1), int2(i2), int3(i3), int4(i4), int5(i5), int6(i6) {}
    // Return operator
    bool operator<(const SixInt &right) const {
      return (
	      int1 < right.int1 ||
	      (int1 == right.int1 && int2 < right.int2) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 < right.int3) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 < right.int4) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 == right.int4 &&
	       int5 < right.int5) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 == right.int4 &&
	       int5 == right.int5 && int6 < right.int6)
	      
	      );
    }
    // maybe this should be private, not sure
    int int1;
    int int2;
    int int3;
    int int4;
    int int5;
    int int6;
  }; // End class SixInt
  
  // The Antenna class: contains a generic antenna function

  class Antenna {
    
  public: 
    //Bare constructor: name only
  Antenna(string Name="antenna") : name(Name),
      humanName(Name) {useFiniteTerms = true; isInitPtr = false; isOnSav = false;};
    
    //Constructor with PDG codes: 
    //  Antenna(string Name,vector<int> IdOld, vector<int> IdNew)
    //  : name(Name), idOld(IdOld), idNew(IdNew), 
    //	isInit(false), {useFiniteTerms=true;};
    
    // Destructor.
    virtual ~Antenna() {};  
    
    // Initialize
    bool init(bool checkAntenna=true);
    void initPtr(ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, 
		 Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    }
    bool setC(int i=0,int j=0,double c=0.0); // Set C[i][j]
    bool setD(int j=0,double d=0.0);         // Set D[j]
    bool setE(int j=0,double e=0.0);         // Set E[j]
    bool setK(double k=0.0);                 // Set K
    // MR 10.7.2010: add MNum and MDenom
    bool setMNum(int, int, int, int, int, int, double);
    bool setMDenom(int, int, int, int, int, int, double);
    // MR 19.7.2010: add CollinearGluonSplitting
    bool setCollinearGluonSplitting(double);
    bool setParm(bool&, string);
    bool setParm(int&, string);
    bool setParm(double&, string);
    bool setParm(string&, string);
    void setHumanName(string humanNameIn) {humanName = humanNameIn;}
    
    // Functions to "get"
    bool isOn() {return (isInit ? isOnSav : false);}
    string getHumanName() { return humanName; }
    string getName() { return name; }
    int getInputMode() {return inputMode;}
    int getEvolutionType() {return evolutionType;};     
    int getKineMapType() {return kineMapType;}  
    double getChargeFactor() {return chargeFactor;}  
    int getCutoffType() { return cutoffType; }                 
    double getCutoffScale() { return cutoffScale; }                 
    
    // The antenna function for this Sudakov
    // MR 7.12.2010: for testing, add a boolean argument to a2 determining 
    // whether the antenna should return numbers >= 0 or not
    double a2ofS(double sa1,double s1b, double s, double mass1=0., double mass2=0.,
					double mass3=0., bool forcepositive=true) {
      return (s != 0) ? a2(sa1/s, s1b/s, mass1/sqrt(s), mass2/sqrt(s), 
			   mass3/sqrt(s), forcepositive) : 0.0;
    };
    
    double a2(double ya1,double y1b, double mu1=0., double mu2=0., 
				double mu3=0., bool forcepositive=true); 
    
    // The finite part polynomial coefficients 
    // (cannot be made static)
    void readC();                      // Read C,D,E,K from Settings
    double getC(int i=0, int j=0);     // Return C[i][j]
    double getD(int j=0);              // Return D[j]
    double getE(int j=0);              // Return E[j]
    double getK();                     // Return K
    // MR 7.10.2010: add MNum, MDenom
    double getMNum(int, int, int, int, int, int);
    double getMDenom(int, int, int, int, int, int);
    // MR 19.7.2010: add CollinearGluonSplitting
    double getCollinearGluonSplitting();
    void clear();                      // Clear all C,D,E,K
    
    // Print a label for this antenna
    void label(ostream& os=cout);
	
    // MR 21.10.2010: check combines IsPositive and IsFreeOfDeadZones
    bool check(int nPointsCheck, double massmu1, double massmu2, 
	       double massmu3, bool dorepair=false);
    
    // MR 16.7.2010: list the coefficients of the antenna
    void listCoefficients(ostream& os=cout);
    
    // MR 21.12.2010: IsPositive checks whether an antenna is positive
    // given a set of masses. It has an optional argument dorepair 
    // (default false) which governs whether the finite coefficients 
    // are changed to make the antenna positive.
    bool IsPositive( int nPointsCheck, double mu1, double mu2, 
		     double mu3, bool dorepair = false );
    // MR 21.12.2010: IsFreeOfDeadZones checks whether an antenna has 
    // dead zones given a set of masses. It has an optional argument 
    // dorepair (default false) which governs whether the finite 
    // coefficients are changed to make the dead zones go away.
    bool IsFreeOfDeadZones( int nPointsCheck, double mu1, 
			    double mu2, double mu3, bool dorepair = false );
    // MR 21.12.2010: rescaleFiniteTerms rescales all the finite terms
    // of an antenna, returning true if it was able to do so, 
    // false otherwise
    bool rescaleFiniteTerms(double rescalingFactor);
    
  private:
    
    // Antenna Parameters 
    int evolutionType;
    int kineMapType;
    double chargeFactor;
    
    // Cutoff
    int cutoffType;
    double cutoffScale;

    // Antenna function name 
    string name;       // in xml/cmnd files (e.g., Vincia:qq:Emit)
    string humanName;  // human readable format, derived.
    
    // In- and Out- particles in PDG code format
    vector<int> idOld;
    vector<int> idNew;
    vector<string> nameOld;
    vector<string> nameNew;
    
    // Coefficients of the antenna function 
    // (There is only one possible input mode for now: Laurent)
    int inputMode;  
    // Coefficients of Laurent series
    map<int, map<int, double> > LaurentC;
    // Additional coefficients for sector antennae
    map<int,double> LaurentD,LaurentE;
    double LaurentK;
    // MR 10.7.2010: six-index coefficients for numerator and
    // denominator of massive antennae, access them by an array
    // of six integers
    map <SixInt, double> LaurentMNum;
    map <SixInt, double> LaurentMDenom;
	// MR 19.7.2010: add CollinearGluonSplitting
	double CollinearGluonSplitting;
    // Flag whether to use finite terms or not
    bool useFiniteTerms;
    // MR 2.3.2011: Flag whether to use mass terms or not (for testing)
    bool useMassTerms;
    // Dead Zone Avoidance Factor
    double deadZoneAvoidance;
    
    // Initialized flag
    bool isInit;

    // On/off flag
    bool isOnSav;
    
    // Pointers to Pythia-8 classes
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
    
    // Verbose setting
    int verbose;

  };

  // AntennaSet = List of Antennae 
  
  class AntennaSet {

  public:
    
    void setAnt(int iAntIn, Antenna antIn) {
      if (iAntIn >= int(antSav.size())) antSav.resize(iAntIn+1);
      antSav[iAntIn]=antIn;
      antIndxSav[antIn.getName()]=iAntIn;
    }
    bool exists(int iAntIn) {
      return (iAntIn<0 || iAntIn >= int(antSav.size())) ? false : true;
    }
    Antenna* getAnt(int iAntIn) {
      return (exists(iAntIn) ? &antSav[iAntIn] : 0);
    }
    string getName(int iAntIn) {
      return (exists(iAntIn) ? antSav[iAntIn].getName() : "none");
    }
    string getHumanName(int iAntIn) {
      return (exists(iAntIn) ? antSav[iAntIn].getHumanName() : "none");
    }
    int getIndx(string antNameIn) {return antIndxSav[antNameIn];}
    
  private:
    
    vector<Antenna> antSav;
    map<string,int> antIndxSav;

  }; // End class AntennaSet

  
} // End Vincia namespace

#endif
