// VinciaAntenna.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Include the VinciaAntenna header file
#include "VinciaAntenna.h"

using namespace Pythia8;

namespace Vincia {
  
  //**********************************************************************
  
  // The Antenna class and its members
  
  //*********
  
  // Initialize Antenna

  bool Antenna::init(bool checkAntenna) {

    // Check that pointers initialized
    if (! isInitPtr) {
      printOut("Antenna::init","Cannot initialize, pointers not set.");
      return false;
    }

    // Innocent until proven guilty
    isInit=true;

    // Force inputMode = 1 (double Laurent rep) for the time being
    inputMode = 1;

    // Set verbose level
    setParm(verbose,"verbose");

    // Read generic in- and out- states
    string o1,o2,n1,n2,n3,n4;
    if (name.find("Vincia:qQ:emit") != string::npos ) {
      o1="q";o2="qbar";n1="q";n2="g";n3="qbar";
    }
    else if (name.find("Vincia:qg:emit") != string::npos ) {
      o1="q";o2="g";n1="q";n2="g";n3="g";
    }
    else if (name.find("Vincia:gg:emit") != string::npos ) {
      o1="g";o2="g";n1="g";n2="g";n3="g";}      
    else if (name.find("Vincia:qg:split") != string::npos ) {
      o1="q";o2="g";n1="q";n2="qbar'";n3="q'";
    }
    else if (name.find("Vincia:gg:split") != string::npos ) {
      o1="g";o2="g";n1="g";n2="qbar";n3="q";
    }
    else {
      if (verbose >= 1)
	printErr(name+":init","unknown generic antenna");
      return false;
    }
    nameOld.push_back(o1);
    nameOld.push_back(o2);
    nameNew.push_back(n1);
    nameNew.push_back(n2);
    nameNew.push_back(n3);
    if (n4 != "") nameNew.push_back(n4);
    for (int ic=0;ic<int(nameOld.size());ic++) {
      if (nameOld[ic]=="q") idOld.push_back(1);
      if (nameOld[ic]=="qbar") idOld.push_back(-1);
      if (nameOld[ic]=="g") idOld.push_back(21);
      if (nameOld[ic]=="gamma") idOld.push_back(22);
    }
    for (int ic=0;ic<int(nameNew.size());ic++) {
      if (nameNew[ic]=="q") idNew.push_back(1);
      if (nameNew[ic]=="qbar") idNew.push_back(-1);
      if (nameNew[ic]=="g") idNew.push_back(21);
      if (nameNew[ic]=="gamma") idNew.push_back(22);
      if (nameNew[ic]=="q'") idNew.push_back(2);
      if (nameNew[ic]=="qbar'") idNew.push_back(-2);
    }
    if (humanName == name) humanName=name.substr(7,name.length());

    // First check whether antenna is 2->3
    if (nameOld.size() != 2 || nameNew.size() != 3) {
      if (verbose >= 1)
	printErr(name+":init","unknown antenna type: "+
		 num2str(int(idOld.size()))+"->"+num2str(int(idNew.size())));
      return false;
    }

    // Check if switched off
    if ( ! settingsPtr->flag(name)) {
        clear();
	isOnSav = false;
	setC(0,0,0.0);
	return true;
    } else {
      isOnSav = true;
    }

    // Set whether finite terms to be used in evolution
    isInit=isInit&&setParm(useFiniteTerms,"useFiniteTerms");
    // MR 2.3.2011: Set whether massive terms are used
    isInit=isInit&&setParm(useMassTerms,"useMassTerms");
    // Initialize parameter to use when checking for dead zones
    isInit=isInit&&setParm(deadZoneAvoidance,"deadZoneAvoidance");
    
    // Check charge conservation
    double sumQ=0.0;
    for (unsigned int ic=0;ic<idOld.size();ic++) 
      sumQ+=particleDataPtr->charge(idOld[ic]);
    for (unsigned int ic=0;ic<idNew.size();ic++) 
      sumQ-=particleDataPtr->charge(idNew[ic]);
    if (abs(sumQ) > TINY) {
      if (verbose >= 1)
	printErr(name+":init","antenna does not conserve charge");
      cout<<"   sumQ = "<<sumQ<<" Antenna ID :"<<idOld[1]<<" "
	  <<idOld[2]<<" -> "
	  <<idNew[1]<<" "<<idNew[2]<<" "<<idNew[3]<<endl;
      return false; 
    }

    // Read in antenna function terms (zero if switched off)    
    readC();  

    // Check that antenna function is positive definite 
    // and free of dead zones, if not rescale the finite terms
    if (checkAntenna) {
      bool changedForPositivity = false;
      bool changedForNoDead = false;
      //    int nPoints = (verbose > 1) ? min(1e4,1.0/sqrt(TINY)) : 100;
      int nPointsCheck = 1000;
      
      // MR 21.12.2010: Hard-code the masses at which the antennae 
      // are tested.
      // masses for a03 (m1,0,m2)
      const double a03_masses_input[][2] = 
	{	
	  {0,0}, {0.01,0}, {0.05,0}, {0.3,0}, {0.5,0}, {0.9,0}, {0.95,0},
	  {0.99,0}, {1,0},
	  {0.01,0.01}, {0.05,0.05}, {0.3,0.3}, {0.45,0.45}, {0.49,0.49},
	  {0.01,0.3}, {0.05,0.3}, {0.1,0.3},
	  {0.01,0.95}, {0.01,0.99}, {0.05,0.9}, {0.05,0.95}, 
	  {0.1,0.8}, {0.1,0.9}, {0.3,0.6}, {0.3,0.7}
	};
      vector< vector<double> > a03_masses;
      for (size_t i=0; i<sizeof(a03_masses_input)/(2*sizeof(double)); i++) {
	vector<double> prov_masses;
	prov_masses.push_back(a03_masses_input[i][0]);
	prov_masses.push_back(0);
	prov_masses.push_back(a03_masses_input[i][1]);
	a03_masses.push_back( prov_masses );
      }
      
      // masses for d03 (m,0,0)
      const double d03_masses_input[] = 
	{0, 0.01, 0.05, 0.1, 0.3, 0.5, 0.7, 0.9, 0.95, 0.99, 1};
      vector< vector<double> > d03_masses;
      for (size_t i=0; i<sizeof(d03_masses_input)/sizeof(double); i++) {
	vector<double> prov_masses;
	prov_masses.push_back(d03_masses_input[i]);
	prov_masses.push_back(0);
	prov_masses.push_back(0);
	d03_masses.push_back( prov_masses );
      }
      
      // masses for e03 (m1,m2,m2)
      const double e03_masses_input[][3] = 
	{	
	  {0,0,0}, {0.01,0,0}, {0.05,0,0}, {0.3,0,0}, {0.5,0,0}, 
	  {0.9,0,0}, {0.95,0,0}, {0.99,0,0}, {1,0,0},
	  {0,0.01,0.01}, {0.01,0.01,0.01}, {0.05,0.01,0.01}, 
	  {0.3,0.01,0.01}, {0.5,0.01,0.01}, {0.9,0.01,0.01}, 
	  {0.95,0.01,0.01}, {0.98,0.01,0.01},
	  {0,0.05,0.05}, {0.01,0.05,0.05}, {0.05,0.05,0.05}, 
	  {0.3,0.05,0.05}, {0.5,0.05,0.05}, {0.85,0.05,0.05}, 
	  {0.9,0.05,0.05}, 
	  {0,0.3,0.3}, {0.01,0.3,0.3}, {0.05,0.3,0.3}, 
	  {0.3,0.3,0.3}, {0.35,0.3,0.3}, {0.4,0.3,0.3}, 
	  {0,0.45,0.45}, {0.01,0.45,0.45}, {0.05,0.45,0.45}, 
	  {0.1,0.45,0.45}
	};
      vector< vector<double> > e03_masses;
      for (size_t i=0; i<sizeof(e03_masses_input)/(3*sizeof(double)); i++) {
	vector<double> prov_masses;
	prov_masses.push_back(e03_masses_input[i][0]);
	prov_masses.push_back(e03_masses_input[i][1]);
	prov_masses.push_back(e03_masses_input[i][2]);
	e03_masses.push_back( prov_masses );
      }
      
      vector< vector<double> > f03_masses;
      vector<double> prov_masses0;
      prov_masses0.push_back(0);
      prov_masses0.push_back(0);
      prov_masses0.push_back(0);
      f03_masses.push_back(prov_masses0);
      
      // masses for g03 (0,m,m)
      const double g03_masses_input[] =
	{ 0., 0.01, 0.05, 0.1, 0.3, 0.45, 0.49, 0.5 };
      vector< vector<double> > g03_masses;
      for (size_t i=0; i<sizeof(g03_masses_input)/sizeof(double); i++) {
	vector<double> prov_masses;
	prov_masses.push_back(0);
	prov_masses.push_back(g03_masses_input[i]);
	prov_masses.push_back(g03_masses_input[i]);
	g03_masses.push_back( prov_masses );
      }
      
      vector< vector< double > > masscheckvalues;
      if (name == "Vincia:qQ:emit") {
	masscheckvalues = a03_masses;
      }
      else if (name == "Vincia:qg:emit") {
	masscheckvalues = d03_masses;
      }
      else if (name == "Vincia:qg:split") {
	masscheckvalues = e03_masses;
      }
      else if (name == "Vincia:gg:emit") {
	masscheckvalues = f03_masses;
	
      }
      else if (name == "Vincia:gg:split") {
	masscheckvalues = g03_masses;
      }
      else {
	// complain if the antenna does not have a known
	// name
	if (verbose>=3) {
	  printOut("init", name + " is not a known antenna name ");
	}
	masscheckvalues = f03_masses;
      }
      
      for (unsigned int i = 0; i < masscheckvalues.size(); ++i) {
	double mu1 = masscheckvalues[i][0];
	double mu2 = masscheckvalues[i][1];
	double mu3 = masscheckvalues[i][2];
	if (verbose >= 4) {
	  printOut(name + ":init", 
		   "checking with reduced masses " +
		   num2str( mu1 ) + ", " +
		   num2str( mu2 ) + ", " +
		   num2str( mu3 ) + ".");
	}
	// if all the positivity tests passed so far, call
	// IsPositive without instructing it to repair
	if (not changedForPositivity) {
	  changedForPositivity = not IsPositive(nPointsCheck, 
						mu1, mu2, mu3, false);
	}
	// if a positivity test failed up to here, call IsPositive
	// instructing it to change the parameters if needed
	if (changedForPositivity) {
	  if (not IsPositive(nPointsCheck,mu1,mu2,mu3,true)) {
	    return false;
	  };
	}
	// the analogue for dead zones
	if (not changedForNoDead) {
	  changedForNoDead = not IsFreeOfDeadZones(
						   nPointsCheck, 
						   mu1, mu2, mu3, false);
	}
	if (changedForNoDead) {
	  if (not IsFreeOfDeadZones(nPointsCheck, 
				    mu1, mu2, mu3, true) ) {
	    return false;
	  };
	}
	
      }
      
      // give a message if coefficients were changed and the 
      // verbosity is high enough
      
      if (changedForPositivity) {
	if (verbose >= 1)
	  printErr(humanName+":init","warning: Finite terms changed to ensure positivity");
	listCoefficients();
      } else if (changedForNoDead) {
	if (verbose >= 1)
	  printErr(humanName+":init","warning: Finite terms changed to ensure absence of dead zones");
	listCoefficients();
      } else {
	if (verbose >= 4)
	  printOut(humanName+":init","ok: positive definite, no dead zones");
      }
    }
    

    // Initialize evolution settings
    isInit=isInit&&setParm(evolutionType,"evolutionType"); 
    isInit=isInit&&setParm(kineMapType,"kineMapType");
    
    // Set chargeFactor (i.e., CA, CF, TR, etc.)
    int showerNLC = settingsPtr->mode("VinciaShower:NLC");
    if (showerNLC == 0) {
      // Read chargeFactor from antennae.
      isInit=isInit&&setParm(chargeFactor,"chargeFactor");
    } else {
      // If strict LC or systematic NLC, start from strict LC
      if (name.find("emit") != string::npos) chargeFactor = CA;
      else chargeFactor = TR;
    }
    
    // IR hadronization cutoff type and scale
    isInit=isInit&&setParm(cutoffType,"cutoffType");
    if (cutoffType == 0) cutoffType=evolutionType;
    isInit=isInit&&setParm(cutoffScale,"cutoffScale");
    
    // Return true or false, depending on how checks went.
    return isInit;
  }
  
//*********

// Set antenna parameters from Settings database

bool Antenna::setParm(bool& par, string parName) {
  bool found=true;
  if (settingsPtr->isFlag(name+":"+parName)) 
    par = settingsPtr->flag(name+":"+parName);
  else if (settingsPtr->isFlag("VinciaShower:"+parName)) 
    par = settingsPtr->flag("VinciaShower:"+parName);
  else if (settingsPtr->isFlag("Vincia:"+parName)) 
    par = settingsPtr->flag("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName); }
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found"); }
  return found;
}

bool Antenna::setParm(int& par, string parName) {
  bool found=true;
  if (settingsPtr->isMode(name+":"+parName)) 
    par = settingsPtr->mode(name+":"+parName);
  else if (settingsPtr->isMode("VinciaShower:"+parName)) 
    par = settingsPtr->mode("VinciaShower:"+parName);
  else if (settingsPtr->isMode("Vincia:"+parName)) 
    par = settingsPtr->mode("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose>= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(double& par, string parName) {
  bool found=true;
  if (settingsPtr->isParm(name+":"+parName)) 
    par = settingsPtr->parm(name+":"+parName);
  else if (settingsPtr->isParm("VinciaShower:"+parName)) 
    par = settingsPtr->parm("VinciaShower:"+parName);
  else if (settingsPtr->isParm("Vincia:"+parName)) 
    par = settingsPtr->parm("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(string& par, string parName) {
  bool found=true;
  if (settingsPtr->isWord(name+":"+parName)) 
    par = settingsPtr->word(name+":"+parName);
  else if (settingsPtr->isWord("VinciaShower:"+parName)) 
    par = settingsPtr->word("VinciaShower:"+parName);
  else if (settingsPtr->isWord("Vincia:"+parName)) 
    par = settingsPtr->word("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

//*********

// Read in Laurent coefficients from Settings database

void Antenna::readC() {
  if (verbose >= 5)
    printOut(humanName+":readC","reading antenna coefficients ");
  // Antenna function
  clear();
 
  // Input as Laurent Series
  if (inputMode == 1) {

    //------------------------------------------------------------------
    // 1st part: read in massless antenna coefficients

    for (int ic=-1;ic<=LaurentIndexMax;ic++) {
      string baseName("LaurentC("+num2str(ic,0)+",");
      for (int j=-1;j<=LaurentIndexMax;j++) {
	string CName(baseName+num2str(j,0)+")"); 
	if (settingsPtr->isParm(name+":"+CName)) {	
	  if (setC(ic,j,settingsPtr->parm(name+":"+CName))) {
	    int verb = (getC(ic,j) == 0.0) ? 8 : 6; 
	    if (verbose >= verb)
	      printOut(humanName+":readC",
		     CName+" = "+num2str(getC(ic,j),6));
	  }
	} 
      } 
      string DName("LaurentD("+num2str(ic,0)+")");
      if (settingsPtr->isParm(name+":"+DName)) {	
	if (setD(ic,settingsPtr->parm(name+":"+DName))) {
	  int verb = (getD(ic) == 0.0) ? 8 : 6; 
	  if (verbose >= verb)
	    printOut(humanName+":readC",
		   DName+" = "+num2str(getD(ic),6));
	}
      } 
      string EName("LaurentE("+num2str(ic,0)+")");
      if (settingsPtr->isParm(name+":"+EName)) {	
	if (setE(ic,settingsPtr->parm(name+":"+EName))) {
	  int verb = (getE(ic) == 0.0) ? 8 : 6; 
	  if (verbose >= verb)
	    printOut(humanName+":readC",
		     EName+" = "+num2str(getE(ic),6));
	}
      } 
    } 
    string KName("LaurentK");
    if (settingsPtr->isParm(name+":"+KName)) {	
      if (setK(settingsPtr->parm(name+":"+KName))) {
	int verb = (getK() == 0.0) ? 8 : 6; 
	if (verbose >= verb)
	  printOut(humanName+":readC",
		   KName+" = "+num2str(getK(),6));
      }
    }

    // MR 14.7.2010: loop only over entries that do indeed exist
    // use getParms
    map<string,Parm> mnums;
    mnums = settingsPtr->getParmMap( name+":LaurentMNum");
    // loop through the map, add the entries
    for (map<string,Parm>::iterator pt=mnums.begin();
	 pt!=mnums.end();++pt){
      // read in the six numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      size_t pos6 = parname.find(",", pos5+1);
      size_t pos7 = parname.find(")", pos6+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int y12ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int y13ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int y23ord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int mu1ord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      int mu2ord = atoi(parname.substr(pos5+1, pos6-pos5-1).c_str());
      int mu3ord = atoi(parname.substr(pos6+1, pos7-pos6-1).c_str());
      setMNum(y12ord,y13ord,y23ord,mu1ord,mu2ord,mu3ord,
	      pt->second.valNow);
      // output
      int verb = (abs(getMNum(y12ord,y13ord,y23ord,mu1ord,mu2ord,mu3ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(humanName+":readMNum",
		 parname+" = "+num2str(getMNum(y12ord,y13ord,y23ord,mu1ord,mu2ord,mu3ord),6));
      }
    }
    // MR 19.7.2010: read in CollinearGluonSplitting
    // this is a global parameter, it cannot have different values for
    // different antennae
    string cgsname("Vincia:CollinearGluonSplitting");
    if ( settingsPtr->isParm(cgsname) ) {	
      if ( setCollinearGluonSplitting(settingsPtr->parm(cgsname)) ) {
	int verb = (getCollinearGluonSplitting() == 0.0) ? 8 : 6; 
	if (verbose >= verb)
	  printOut(humanName+":readC",
		   cgsname+" = "+num2str(getCollinearGluonSplitting(),6));
      }
    }
    
  } else {
    // Enter alternative input forms here
  }
}
  
//*********

// Set a Laurent coefficient directly

bool Antenna::setC(int ic,int j, double cIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    if (ic < -1 or j < -1) {
      if (verbose >= 1)
	printErr(humanName+":setC",
		 "leading term is (-1,-1) : ignoring entry ("
		 +num2str(ic)+","+num2str(j)+")");
      return false;
    }
    else if (cIn == 0) {
      if (LaurentC.find(ic) != LaurentC.end()) {
	if (LaurentC[ic].find(j) != LaurentC[ic].end()) LaurentC[ic].erase(j);
      } 
    }
    else if (ic == -1 && j == -1 && cIn>0) LaurentC[ic][j]=cIn;
    else if (j == -1 || ic == -1) LaurentC[ic][j]=cIn;
    else LaurentC[ic][j]=cIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1)
      printErr(humanName+":setC",
	     "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

bool Antenna::setD(int j, double dIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    if (j < -1) {
      if (verbose >= 1)
	printErr(humanName+":setD",
	       "leading term is (-1) : ignoring entry ("
	       +num2str(j)+")");
      return false;
    }
    else if (dIn == 0) {
      if (LaurentD.find(j) != LaurentD.end()) LaurentD.erase(j);
    }
    else LaurentD[j]=dIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1) 
      printErr(humanName+":setD",
	       "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

bool Antenna::setE(int j, double eIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    if (j < -1) {
      if (verbose >= 1)
	printErr(humanName+":setE",
		 "leading term is (-1) : ignoring entry ("
		 +num2str(j)+")");
      return false;
    }
    else if (eIn == 0) {
      if (LaurentE.find(j) != LaurentE.end()) LaurentE.erase(j);
    }
    else LaurentE[j]=eIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1) 
      printErr(humanName+":setE",
	       "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

bool Antenna::setK(double kIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    LaurentK=kIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1) 
      printErr(humanName+":setK",
	       "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}


// MR 10.7.2010: analogue to setC for MNum and MDenom
bool Antenna::setMNum(int y12ord, int y13ord, int y23ord, int mu1ord,
						int mu2ord, int mu3ord, double coeffIn) {
  if (inputMode == 1) {
    // Input as Laurent series
	// don't check the range at all
	SixInt index = SixInt(y12ord, y13ord, y23ord, mu1ord, mu2ord, mu3ord);
	// define an iterator
	map <SixInt, double>::iterator it = LaurentMNum.find(index);
    if (abs(coeffIn) <= TINY) {
	  // if the entry the absolute value of which is set <= TINY does 
	  // exist, remove it from the map to increase performance 
      if (it != LaurentMNum.end()) {
		LaurentMNum.erase( it );
      } 
    }
	else LaurentMNum[index] = coeffIn;
  } 
  else {
    // Enter alternative input modes here
    if (verbose >= 1)
      printErr(humanName+":setMNum",
	     "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

bool Antenna::setMDenom(int y12ord, int y13ord, int y23ord, int mu1ord,
						int mu2ord, int mu3ord, double coeffIn) {
  if (inputMode == 1) {
    // Input as Laurent series
	// don't check the range at all
	SixInt index = SixInt(y12ord, y13ord, y23ord, mu1ord, mu2ord, mu3ord);
	// define an iterator
	map <SixInt, double>::iterator it = LaurentMDenom.find(index);
    if ( abs(coeffIn) <= TINY ) {
	  // if the entry the absolute value of which is set <= TINY does 
	  // exist, remove it from the map to increase performance 
      if (it != LaurentMDenom.end()) {
		LaurentMDenom.erase( it );
      } 
    }
    else LaurentMDenom.insert(make_pair(index, coeffIn));
  } 
  else {
    // Enter alternative input modes here
    if (verbose >= 1)
      printErr(humanName+":setMDenom",
	     "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

  // MR 19.7.2010: CollinearGluonSplitting
  bool Antenna::setCollinearGluonSplitting(double cgsin) {
    if (inputMode == 1) {
      // Input as Laurent series
      CollinearGluonSplitting=cgsin;
    } else {
      // Enter alternative input modes here
      if (verbose >= 1) 
	printErr(humanName+":setCollinearGluonSplitting",
		 "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  

  
//*********

// Function to "get" value of a Laurent coefficient

double Antenna::getC(int ic,int j) {     
  if (inputMode == 1) {
    // Input as Laurent series
    map<int,map<int,double> >::iterator it;
    map<int,double>::iterator jt;
    it=LaurentC.find(ic);
    if (it==LaurentC.end()) return 0.0;
    jt=LaurentC[ic].find(j);
    if (jt==LaurentC[ic].end()) return 0.0;
    return LaurentC[ic][j];
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
double Antenna::getD(int j) {     
  if (inputMode == 1) {
    // Input as Laurent series
    map<int,double>::iterator jt;
    jt=LaurentD.find(j);
    if (jt==LaurentD.end()) return 0.0;
    return LaurentD[j];
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
double Antenna::getE(int j) {     
  if (inputMode == 1) {
    // Input as Laurent series
    map<int,double>::iterator jt;
    jt=LaurentE.find(j);
    if (jt==LaurentE.end()) return 0.0;
    return LaurentE[j];
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
double Antenna::getK() {       
  if (inputMode == 1) {
    return LaurentK;
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}

// MR 10.7.2010: analogue to getC for MNum and MDenom
double Antenna::getMNum(int y12ord,int y13ord, int y23ord, int mu1ord,
						int mu2ord, int mu3ord) {     
  if (inputMode == 1) {
    // Input as Laurent series
	SixInt index = SixInt(y12ord, y13ord, y23ord, mu1ord, mu2ord, mu3ord);
    map <SixInt, double>::iterator it = LaurentMNum.find( index );
	// return zero for non-existing entries
    if (it==LaurentMNum.end()) return 0.0;
    return it->second;
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}  
double Antenna::getMDenom(int y12ord,int y13ord, int y23ord, int mu1ord,
						int mu2ord, int mu3ord) {     
  if (inputMode == 1) {
    // Input as Laurent series
	SixInt index = SixInt(y12ord, y13ord, y23ord, mu1ord, mu2ord, mu3ord);
    map <SixInt, double>::iterator it = LaurentMDenom.find( index );
	// return zero for non-existing entries
    if (it==LaurentMDenom.end()) return 0.0;
    return it->second;
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}

double Antenna::getCollinearGluonSplitting() {       
  if (inputMode == 1) {
    return CollinearGluonSplitting;
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}

//*********

// Zero all Laurent coefficients

void Antenna::clear() {
  if (inputMode == 1) {
    // Clear Laurent series coeffiecents
    LaurentC.clear(); 
    LaurentD.clear(); 
    LaurentE.clear();
    LaurentK=0.0;
	// MR 11.7.10: include MNum, MDenom
	LaurentMNum.clear();
	LaurentMDenom.clear();
	// MR 19.7.2010: include CollinearGluonSplitting
	CollinearGluonSplitting=0;
  } else {
    // Enter alternative input modes here
  }
}

//*********

// The antenna function 
// MR 9.7.2010: change to accept massmu1, massmu2, massmu3 as well
// MR 7.12.2010: added parameter forcepositive (default true) which
// governs whether the antenna is restricted to return numbers >=0
// MR 21.12.2010: The antenna function does always return 0 if the 
// arguments are outside the physical phase space or too close to 
// its boundary 

double Antenna::a2(double y12, double y23, double massmu1, double massmu2,
					double massmu3, bool forcepositive) {

  // First check if switched off
  if ( !isOn() ) return 0.0;
  
  // initialize the return value
  double val=0.0;
  // calculate the missing invariant
  double y13 =
    1 - y12 - y23 - pow2( massmu1 ) - pow2( massmu2 ) - pow2( massmu3 );
  // complain if y13 is below a value approaching machine precision
  if (y13 - 2*massmu1*massmu3 < -TINY) {
    if (verbose>=3) {
      printOut("VinciaAntenna::a2", "y13 < 2*mu1*mu3 occurred");
      cout << "   y12 = " << y12 ;
      cout << " y23 = " << y23 ;
      cout << " y13 = " << y13 << "\n";
      cout << "   mu1 = " << massmu1 ;
      cout << " mu2 = " << massmu2 ;
      cout << " mu3 = " << massmu3 <<endl;      
    }
    return 0.0;
  } 
  // Tolerate values slightly below y13min (due to rounding), 
  // but set them = 2*massmu1*massmu3
  else if (y13 < 2*massmu1*massmu3) {
    y13 = 2*massmu1*massmu3;
  }
  
  // MR 21.12.2010: Use Gram determinant to check whether we are too
  // close to the boundary of the phase space.
  // Check for gDet / Ecm^6 < TINY^3 because the Gram determinant has 
  // mass dimension 6
  // have to check y12min <= y12 <= y12max first because
  // for y12min > y12max, a region with positive gram
  // determinant opens up
  double y12min = 2*massmu1*massmu2;
  double y12max = pow2( 1-massmu3 ) - pow2(massmu1) - pow2(massmu2);
  if ( (y12 - y12min < 0.5*TINY ) || (y12max - y12 < 0.5*TINY) ) {
    if (verbose >= 3) {
      printOut(name+":a2",
	       "y12 = " + num2str(y12, 4) + " outside physical phase "
	       + "space or too close to boundaries y12min = " 
	       + num2str(y12min, 4) + ", y12max = " 
	       + num2str(y12max, 4) + "}, return 0.");
    }
    return 0.0;
  }
  double gdetdimless = gDet(y12, y23, y13, massmu1, massmu2, massmu3);
  if (gdetdimless< pow3(TINY)) {
    if (verbose >= 3) {
      printOut(name+":a2", 
	       "Gram determinant " + num2str(gdetdimless, 15) 
	       + " < " + num2str(pow3(TINY), 15)  + ", return 0\n"
	       + "\ty12 = " + num2str(y12, 15) + ",\n"
	       + "\ty23 = " + num2str(y23, 15) + ",\n"
	       + "\ty13 = " + num2str(y13, 15) + ",\n"
	       + "\tmu1 = " + num2str(massmu1, 15) + ",\n"
	       + "\tmu2 = " + num2str(massmu2, 15) + ",\n"
	       + "\tmu3 = " + num2str(massmu3, 15)
	       );
    }
    return 0.0;
  }
  
  // the coefficients of the massless antennae are always used
  // Input as Laurent series
  if (inputMode == 1) {
    map<int,map<int,double> >::iterator it;
    map<int,double>::iterator jt;
    // Add C(i,j) terms
    for (it=LaurentC.begin();it!=LaurentC.end();++it) {
      int ic=it->first;
      for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
	int j=jt->first;  
	// Only add finite terms if they are "on"
	if (ic < 0 || j < 0 || useFiniteTerms) {
	  double c=jt->second;      
	  val += c * pow(y12,ic) * pow(y23,j);
	}
      }      
    }
    // Add D(j) terms
    for (jt=LaurentD.begin();jt!=LaurentD.end();++jt) {
      int j=jt->first;
      double d=jt->second;
      val += d * pow(y23,j) / (1-y12) ;
    }
    // Add E(j) terms
    for (jt=LaurentE.begin();jt!=LaurentE.end();++jt) {
      int j=jt->first;
      double e=jt->second;
      val += e * pow(y12,j) / (1-y23) ;
    }
    // Add K term
    if (LaurentK != 0) {
      val += LaurentK / (1-y12) / (1-y23) ;
    }
    // MR 19.7.2010: add collinear gluon splitting parameter for gg:emit and
    // qg:emit
    if (CollinearGluonSplitting != 0) {
      if (name == "Vincia:qg:emit") {
	val += CollinearGluonSplitting*(y12-y13)/y23;
      }
      else if (name == "Vincia:gg:emit") {
	val += CollinearGluonSplitting*( (y12-y13)/y23 + (y23-y13)/y12 );
      }
      
    }
    // mass terms
    // MR 2.3.2011: implement the option of turning off mass terms for
    // each antenna separately. This does not make any sense in terms 
    // of physics, but it could be useful for testing    
    if (
      (massmu1 >= TINY || massmu2 >= TINY || massmu3 >= TINY) &&
      useMassTerms
    ) {
      // MR 11.7.2010: write this in analogy to the massless case
      
      // MR 14.6.2010: hard-code the massive denominator and the massive
      // terms contributing to the singularity individually for every
      // antenna
      // TODO: checks for masslessness and mass equality in the
      // antennae
      // initialize the terms
      double denom = 1;
      double numeratormassivesingular = 0;
      double numeratormassivefinite = 0;
      double masslesscontribution = val;
      
      if (name == "Vincia:qQ:emit") {
	denom = 4*massmu1*massmu3 + y12 + y13 + y23;
	numeratormassivesingular =
	  (-8*pow3(massmu1)*massmu3)/pow2(y12) - (4*pow2(massmu1))/y12 - 
	  (2*pow2(massmu3))/y12 - 
	  (2*pow2(massmu1)*y13)/pow2(y12) - 
	  (8*massmu1*pow3(massmu3))/pow2(y23) - 
	  (2*pow2(massmu3)*y12)/pow2(y23) - 
	  (2*pow2(massmu3)*y13)/pow2(y23) - 
	  (2*pow2(massmu1))/y23 - (4*pow2(massmu3))/y23 - 
	  (2*pow4(massmu1))/(y12*y23) - 
	  (4*pow2(massmu1)*pow2(massmu3))/(y12*y23) - 
	  (2*pow4(massmu3))/(y12*y23) - 
	  (4*pow2(massmu1)*y13)/(y12*y23) + 
	  (8*massmu1*massmu3*y13)/(y12*y23) - 
	  (4*pow2(massmu3)*y13)/(y12*y23) - 
	  (2*pow2(massmu1)*y23)/pow2(y12);
	masslesscontribution = masslesscontribution/denom;
      }
      else if (name == "Vincia:qg:emit") {
	denom = pow2(y12 + y13 + y23);
	numeratormassivesingular =
	  -6*pow2(massmu1) - (4*pow4(massmu1))/y12 - 
	  (12*pow2(massmu1)*y13)/y12 - 
	  (2*pow2(massmu1)*pow2(y13))/pow2(y12) - 
	  (4*pow4(massmu1))/y23 - (2*pow6(massmu1))/(y12*y23) - 
	  (3*pow2(massmu1)*y12)/y23 - (8*pow2(massmu1)*y13)/y23 - 
	  (6*pow4(massmu1)*y13)/(y12*y23) - 
	  (6*pow2(massmu1)*pow2(y13))/(y12*y23) - 
	  (7*pow2(massmu1)*y23)/y12 - 
	  (4*pow2(massmu1)*y13*y23)/pow2(y12) - 
	  (2*pow2(massmu1)*pow2(y23))/pow2(y12)
	  // need to correct the collinear gluon splitting
	  -CollinearGluonSplitting*pow2(massmu1)*(y12-y13)/y23;
	masslesscontribution = masslesscontribution/denom;
      }
      else if (name == "Vincia:qg:split") {
	// for the gluon splitting antennae the gluon propagator
	// is taken to belong to the denominator
	denom = pow2(1-pow2(massmu1))*pow2(y23+2*pow2(massmu2));
	numeratormassivesingular =
	  2*pow2(massmu2) - 4*pow2(massmu1)*pow2(massmu2) + 
	  2*pow4(massmu1)*pow2(massmu2) - 
	  2*pow2(massmu2)*y12 + 
	  2*pow2(massmu1)*pow2(massmu2)*y12 + 
	  2*pow2(massmu2)*pow2(y12) - pow2(massmu1)*y23 + 
	  (pow4(massmu1)*y23)/2. + pow2(massmu1)*y12*y23;
	// the coefficients in the massless matrix element are
	// for an overall y23^2 from the gluon propagator,
	// have to multiply by y23^2 to compensate for the 
	// (y23+massmu2^2)^2 in the massive denominator
	masslesscontribution =
	  masslesscontribution*pow2(y23)/denom;
      }
      else if (name == "Vincia:gg:emit") {
      }
      else if (name == "Vincia:gg:split") {
	// the two-particle matrix element is Ecm^4
	denom = 1*pow2(y23+2*pow2(massmu2));
	numeratormassivesingular =
	  2*pow2(massmu2) - 2*pow2(massmu2)*y12 + 
	  2*pow2(massmu2)*pow2(y12);
	// the coefficients in the massless matrix element are
	// for an overall y23^2 from the gluon propagator,
	// have to multiply by y23^2 to compensate for the 
	// (y23+massmu2^2)^2 in the massive denominator
	masslesscontribution =
	  masslesscontribution*pow2(y23)/denom;
      }
      else {
	// complain and return 0 if the antenna does not have a known
	// name
	if (verbose>=3) {
	  printOut(name+":a2", name + " is not a known antenna name ");
	}
	return 0;
      }
      // build up the finite mass-dependent contributions if finite terms
      // are turned on
      if (useFiniteTerms) {
	for (map< SixInt, double>::iterator pt=LaurentMNum.begin();
	     pt!=LaurentMNum.end();++pt) {
	  int y12ord = pt->first.int1;
	  int y13ord = pt->first.int2;
	  int y23ord = pt->first.int3;
	  int mu1ord = pt->first.int4;
	  int mu2ord = pt->first.int5;
	  int mu3ord = pt->first.int6;
	  double presentcoeff = pt->second;
	  numeratormassivefinite += 
	    presentcoeff * pow(y12, y12ord) * pow(y13, y13ord)*
	    pow(y23, y23ord) * pow(massmu1, mu1ord)*
	    pow(massmu2, mu2ord) * pow(massmu3, mu3ord);
	}
      }
      // add up the terms
      if (verbose>=8) {
	printOut(name+":a2","massless antenna = " + num2str(val));
      }
      val = masslesscontribution + numeratormassivesingular/denom
	+ numeratormassivefinite/denom;
      if (verbose>=8) {
	printOut(name+":a2","massive antenna =  " + num2str(val));
      }
      // check that the antenna function is positive
      if (val <= 0.0 && forcepositive) {
	if (verbose>=3 && val <= -TINY) {
	  // have checked further up that I am inside the 
	  // physical phase space
	  printOut("a2","negative value " + num2str(val) 
		   + " for massive antenna function");
	  this->listCoefficients();
	  cout << "massless contribution = " 
	       << masslesscontribution << "\n";
	  cout << "massive singular contribution = " 
	       << numeratormassivesingular/denom << "\n";
	  cout << "massive finite contribution = " 
	       << numeratormassivefinite/denom << "\n";
	  cout << "denom = " << denom << "\n";
	  cout << "y12 = " << y12 << "\n";
	  cout << "y13 = " << y13 << "\n";
	  cout << "y23 = " << y23 << "\n";
	  cout << "mu1 = " << massmu1 << "\n";
	  cout << "mu2 = " << massmu2 << "\n";
	  cout << "mu3 = " << massmu3 << "\n";
	}
	return 0.0;
      }
    } // closing of "if( massmu1 >= TINY ..."
  } // closing of "if InputMode == 1 ..."
  
  // if InputMode is not 1
  else {
    // Enter alternative input forms here
  }
  
  // Return val
  return val;
  
  
  
  
  /*
    double numerator = 0;
    double denom = 0;
    
    // Sum up MNum terms
    for (map< SixInt, double>::iterator pt=LaurentMNum.begin();
    pt!=LaurentMNum.end();++pt) {
    int y12ord = pt->first.int1;
    int y13ord = pt->first.int2;
    int y23ord = pt->first.int3;
    int mu1ord = pt->first.int4;
    int mu2ord = pt->first.int5;
    int mu3ord = pt->first.int6;
    double presentcoeff = pt->second;
    numerator += presentcoeff * pow(y12, y12ord) * pow(y13, y13ord)*
    pow(y23, y23ord) * pow(massmu1, mu1ord)*
    pow(massmu2, mu2ord) * pow(massmu3, mu3ord);
    }
    // sum up MDenom terms
    for (map< SixInt, double>::iterator pt=LaurentMDenom.begin();
    pt!=LaurentMDenom.end();++pt) {
    int y12ord = pt->first.int1;
    int y13ord = pt->first.int2;
    int y23ord = pt->first.int3;
    int mu1ord = pt->first.int4;
    int mu2ord = pt->first.int5;
    int mu3ord = pt->first.int6;
    double presentcoeff = pt->second;
    denom += presentcoeff * pow(y12, y12ord) * pow(y13, y13ord)
    * pow(y23, y23ord) * pow(massmu1, mu1ord)
    * pow(massmu2, mu2ord) * pow(massmu3, mu3ord);
    }
    // check whether the denominator is too small
    if ( abs(denom) <= TINY ) {
    if (verbose >= 3) {
    printOut("a2", string("encountered small denominator ")
    + string("in massive antenna, use massless ")
    + string("antenna instead, input values ")
    + string("were y12 = ") + num2str(y12)
    + ", y13 = " + num2str(y13)
    + ", y23 = " + num2str(y23)
    + ", massmu1 = " + num2str(massmu1)
    + ", massmu2 = " + num2str(massmu2)
    + ", massmu3 = " + num2str(massmu3) );
    }
    // return massless antenna
    val = this->a2(y12, y23, 0, 0, 0);
    }
    // if the denominator is nonzero, check for negative antenna function
    else if ( numerator/denom <= TINY ) {
    if (verbose >= 3) {
    printOut("a2", string("encountered value <= TINY ")
    + string("in massive antenna, use massless ")
    + string("antenna instead, input values ")
    + string("were y12 = ") + num2str(y12)
    + ", y13 = " + num2str(y13)
    + ", y23 = " + num2str(y23)
    + ", massmu1 = " + num2str(massmu1)
    + ", massmu2 = " + num2str(massmu2)
						+ ", massmu3 = " + num2str(massmu3) );
		}
		// return massless antenna
		val = this->a2(y12, y23, 0, 0, 0);
	}
	// if the denominator is not small and the massive antenna is
	// >= TINY, return it
	// if the verbosity is high, output a comparison of
	// the massless and the massive antenna
	val = numerator/denom;
	if (verbose>=6) {
		printOut("a2", "massive ant = " + num2str(val)
						+ " massless ant = "
						+ num2str( this->a2(y12, y23, 0, 0, 0) ) );
	}
	*/
	
	/*
	// find out what the antenna type is
	// quark-gluon antenna
	if (this->name == "Vincia:qg:emit") {
		// TODO: check for nonzero gluon masses
		// the antenna down here is
		// d03massive[y12, y13, y23, massmu1, 0, 3/2, 0]
		double y13 =
			1. - y12 - y23 - pow2(massmu1) - pow2(massmu2) - pow2(massmu3);
		val = 1/pow2(y12+y13+y23)*(
		  massmu1 + 4*y12 + 5*y13 + (4*pow2(y13))/y12 + (3*pow2(y12))/(2*y23) +
          (3*y12*y13)/y23 + (5*pow2(y13))/(2*y23) + (2*pow3(y13))/(y12*y23)
		  + 3*y23 - (pow3(massmu1)*y23)/(y12*y13) + (pow4(massmu1)*y23)/(y12*y13) +
          (3*y13*y23)/y12 + pow2(y23)/y12 + pow2(massmu1)*(-3 - (4*y13)/y12 -
          (2*pow2(y13))/pow2(y12) - (6*y23)/y12 - (4*y13*y23)/pow2(y12) -
		  (2*pow2(y23))/pow2(y12) - pow2(y23)/(y12*y13))
		  );
		// if verbosity is high, state that the massive antenna has been
		// called and compare the massive with the massless antenna
		if (verbose >= 6) {
		  printOut("massive quark-gluon antenna",
					"massive ant ="+ num2str(val) + ", massless "
					+"ant=" + num2str(this->a2(y12, y23, 0, 0, 0)));

		}
	}
	// quark-antiquark antenna
	else if (this->name == "Vincia:qQ:emit") {
		// the antenna below is
		// A03massive[y12, y13, y23, massmu1, massmu3, 0]
		// TODO: check for nonzero gluon mass
		double y13 =
			1. - y12 - y23 - pow2(massmu1) - pow2(massmu2) - pow2(massmu3);
	    val =
		  2/(4*massmu1*massmu3 + y12 + y13 + y23)*(
		    (-4*pow3(massmu1)*massmu3)/pow2(y12) + y13/y12 +
			pow2(massmu3)*(-(y12/pow2(y23)) - y13/pow2(y23) - (1/y23)) +
			massmu1*((-4*pow3(massmu3))/pow2(y23) + (4*massmu3*y13)/(y12*y23)) + 
			y12/(2*y23) + y13/y23 + pow2(y13)/(y12*y23) + y23/(2*y12) +
			pow2(massmu1)*(-(1/y12) - y13/pow2(y12) - y23/pow2(y12))
		  );
		// if verbosity is high, state that the massive antenna has been
		// called and compare the massive with the massless antenna
		if (verbose >= 6) {
		  printOut("massive quark-antiquark antenna",
					"massive ant ="+ num2str(val) +
					" mu1 =" + num2str(massmu1) +
					" mu3 =" + num2str(massmu3) +
					", massless ant=" + num2str(this->a2(y12,y23,0,0,0)));
		}

	}
	else {
		// complain
		if (verbose >= 3) {
			printOut("Antenna","only d03 and A03 massive implemented, returning massless antenna");
		}
		val = this->a2(y12,y23,0,0,0);
	}
	*/
  }
  
// MR 21.12.2010: Changed the check function to rely on IsPositive
// and IsFreeOfDeadZones. This loses a factor of 2 in efficiency
// (the same phase space point is checked twice, once by IsPositive,
// once by IsFreeOfDeadZones), but it makes the function easier to read.

bool Antenna::check(int nPointsCheck, double mu1, double mu2, 
						double mu3, bool dorepair) {
  bool isPositive = 
	this->IsPositive(nPointsCheck, mu1, mu2, mu3, dorepair);
  bool isFreeOfDeadZones = 
	this->IsFreeOfDeadZones(nPointsCheck, mu2, mu2, mu3, dorepair);
  return (isPositive && isFreeOfDeadZones);
}
  
/*
// MR 15.7.2010: made the postivity test for the massive phase space
// its own function since it is going to be called for many mass
// configurations

bool Antenna::check(int nPointsCheck, double massmu1,
									double massmu2, double massmu3){
	// initialize the flags
	bool isPos = true;
	bool noDead = true;
	// calculate the phase space boundaries
	double y12min = 2*massmu1*massmu2;
	double y12max = pow2(1 - massmu3) - pow2(massmu1) - pow2(massmu2);
	for (int i=1;i<nPointsCheck;i++) {
      // Generate a random point in PS
	  // generate y12 flat between minimum and maximum, then
	  // y23 flat between y23min(y12) and y23max(y12), this
	  // results in an oversampling where the y23-range is small
      double y12= y12min + rndmPtr->flat()*(y12max-y12min);
	  double y23minofy12=
			-(-2*pow2(massmu2) + 2*pow2(massmu1)*
			pow2(massmu2) + 2*pow4(massmu2) + 
			2*pow2(massmu2)*pow2(massmu3) - y12 + 
			pow2(massmu1)*y12 + 3*pow2(massmu2)*y12 + 
			pow2(massmu3)*y12 + pow2(y12) + 
			sqrt(1 - pow2(massmu1) - pow2(massmu2) - 
			2*massmu3 + pow2(massmu3) - y12)*
			sqrt(1 - pow2(massmu1) - pow2(massmu2) + 
			2*massmu3 + pow2(massmu3) - y12)*
			sqrt(-4*pow2(massmu1)*pow2(massmu2) + 
			pow2(y12)))/
			(2.*(pow2(massmu1) + pow2(massmu2) + y12));
	  double y23maxofy12 = 
			-(-2*pow2(massmu2) + 2*pow2(massmu1)*
			pow2(massmu2) + 2*pow4(massmu2) + 
			2*pow2(massmu2)*pow2(massmu3) - y12 + 
			pow2(massmu1)*y12 + 3*pow2(massmu2)*y12 + 
			pow2(massmu3)*y12 + pow2(y12) - 
			sqrt(1 - pow2(massmu1) - pow2(massmu2) - 
			2*massmu3 + pow2(massmu3) - y12)*
			sqrt(1 - pow2(massmu1) - pow2(massmu2) + 
			2*massmu3 + pow2(massmu3) - y12)*
			sqrt(-4*pow2(massmu1)*pow2(massmu2) + 
			pow2(y12)))/
			(2.*(pow2(massmu1) + pow2(massmu2) + y12));
	  double y23 =
		y23minofy12 + rndmPtr->flat()*(y23maxofy12-y23minofy12);

      // Evaluate a2 with and without finite terms, tell a2 to allow
	  // negative return values
      double aFull   = a2(y12,y23, massmu1, massmu2, massmu3, false);
      bool useFinSav = useFiniteTerms;
	  // Don't want a2 to complain here if the antenna is negative
	  int verboseSav = verbose;
      useFiniteTerms = false;
	  verbose		 = -1;
      double aSing   = a2(y12,y23, massmu1, massmu2, massmu3, false);
      useFiniteTerms = useFinSav;
	  verbose		 = verboseSav;
	  
	  // check for negative values or dead zones
	  if ((aFull < - TINY) || (aFull < deadZoneAvoidanceFactor * aSing)) {
	    if (verbose <= 3) {
			printOut( name+":check", 
						string("negative value or dead zone found") +
						string(",\n y12 = ") + num2str(y12) +
						string(",\n y23 = ") + num2str(y23) +
						string(",\n mu1 = ") + num2str(massmu1) +
						string(",\n mu2 = ") + num2str(massmu2) +
						string(",\n mu3 = ") + num2str(massmu3) +
						string(",\n aFull = ") + num2str(aFull) +
						string(",\n aSing = ") + num2str(aSing) +
						string(",\n deadZoneAvoidanceFactor = ") + 
							num2str(deadZoneAvoidanceFactor) +
						string(",\n antenna coefficients are:\n")
			);
			this->listCoefficients();
		}
		// check the singular part
		// if aSing is negative, try to lower the collinear gluon
	    // splitting parameter
		  if (aSing == 0) {
			  // if the collinear gluon splitting parameter is zero, 
			  // something is very wrong
			  if (CollinearGluonSplitting <= TINY) {
					if (verbose >= 0) { 
						printErr(name+":check",
						"error: singular term for y12 = " + num2str(y12)
						+ ", y23 = " + num2str(y23) + ", mu1 = "
						+ num2str(massmu1) + ", mu2 = " + num2str(massmu2)
						+ ", mu3 = " + num2str(massmu3) + " has negative value "
						+ num2str(aSing));
						return false;
					}

			  }
			  // if it is not yet zero, lower the gluon splitting parameter
			  // by 0.1
			  else {
					if (verbose >= 5) {
						printOut(name+":check",
									"negative value of the antenna function with "
									+ string("finite parts turned off, ") 
									+ string("reducing CollinearGluonSplitting"));
						printOut(name+":check", "old parameter values:");
						listCoefficients();
					}
					double oldCollinearGluonSplitting =
						getCollinearGluonSplitting();
					setCollinearGluonSplitting(
						max(0., oldCollinearGluonSplitting-0.1 ) );
			  }
		  }

		// if the singular part is positive, loop through the finite terms 
		// of the antenna and multiply all of them by 0.9
		else {
			if (inputMode == 1) {
				// list the old parameter values
				if (verbose >= 5) {
					printOut(name+":check", "antenna function with finite "
								+ string("terms negative or with dead zone, ")
								+ string("reducing magnitude of finite terms."));
					printOut(name+":check", "old parameter values:");
					listCoefficients();
				}
				// scale down C(i,j) terms
				map<int,map<int,double> >::iterator it;
				map<int,double>::iterator jt;
				for (it=LaurentC.begin();it!=LaurentC.end();++it) {
				int ic=it->first;
					for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
						int jc=jt->first;  
						if ( ic >= 0 && jc >= 0) {
							double oldcoeff = jt->second;
							double newabs =
								min( 0.9*abs(oldcoeff), abs(oldcoeff) - 0.1 );
							// round down to the closest multiple of 0.1
							newabs = floor( newabs*10 )/10;
							setC(ic, jc,oldcoeff*newabs/abs(oldcoeff) );
						}
					}
				} // end of loop through C(i,j) terms
				// scale down the MNum terms
				for (map< SixInt, double>::iterator pt=LaurentMNum.begin();
					pt!=LaurentMNum.end();++pt) {
					int y12ord = pt->first.int1;
					int y13ord = pt->first.int2;
					int y23ord = pt->first.int3;
					int mu1ord = pt->first.int4;
					int mu2ord = pt->first.int5;
					int mu3ord = pt->first.int6;
					double oldcoeff = pt->second;
					double newabs =
						min( 0.9*abs(oldcoeff), abs(oldcoeff) - 0.1 );
					// round down to the closest multiple of 0.1
					newabs = floor( newabs*10 )/10;
					setMNum(y12ord, y13ord, y23ord, mu1ord, mu2ord, mu3ord,
								oldcoeff*newabs/abs( oldcoeff ) );
					
				}
				// no other terms included in scaling down at the moment
			} // end of if (inputMode == 1)
			else {
				if (verbose >= 3) {
					printOut("check", "inputMode " + num2str( inputMode )
							+ " not recognized.");
				}
				return false;
			}
		}
		// set isPos = false and/or noDead = False
		if (aFull == 0 ) {
			isPos = false;
		}
		if (aFull < deadZoneAvoidanceFactor * aSing) {
			noDead = false;
		}
		// list the new parameters
		if (verbose >= 5) {
			printOut(name+":check", "new parameter values:");
			listCoefficients();
		}
		
		// reset the loop
		i = 1;
	  } // end of if (aFull < - TINY || aFull < ... )

    } // closing bracket of for loop throwing phase space points
  // if no negative values were found, give message and return true
  if ((not isPos) || (not noDead) ) {
	  if (not isPos) {
		  if (verbose >= 1)
			printErr(humanName+":check","warning: parameters changed to "
								+ string("ensure positivity.") );
	  }
	  else if (not noDead) {
		if (verbose >= 1)
			printErr(humanName+":check","warning: parameters changed to "
							+ string("ensure absence of dead zones.") );
	  }
	  // list the coefficients of the antenna if anything was changed
	  if (verbose >= 1) listCoefficients();
  }
  else if (verbose >= 4) {
      printOut(humanName+":check","ok: positive definite, no dead zones");
  }   
  // if the program arrives here, it has checked and, if necessary, repaired
  // the antenna
  return true;
}
*/
// MR 21.12.2010: IsPositive checks whether an antenna is positive
// given a set of masses. It has an optional argument dorepair (default 
// false) which governs whether the finite coefficients are changed 
// to make the antenna positive.
bool Antenna::IsPositive( int nPointsCheck, double mu1, double mu2, 
								double mu3, bool dorepair ) {
  // initialize the flags
  bool isPos = true;
  
  // don't want the antenna to complain, set verbosity to -1
  int verboseSav = verbose;
  verbose = -1;
  // calculate the phase space boundaries
  double y12min = 2*mu1*mu2;
  double y12max = pow2(1 - mu3) - pow2(mu1) - pow2(mu2);
  // loop through the phase space
  int numberofsteps = floor( sqrt( nPointsCheck ) );
  double y12StepWidth = (y12max - y12min)/numberofsteps;
  // i = 0 and i <= numberofsteps are on the boundary of the phase
  // space, omit them
  for (int i=1; i<=numberofsteps-1; i++) {
    double y12 = y12min + i*y12StepWidth;
    double y23min = y23tildeMin(y12, mu1, mu2, mu3);
    double y23max = y23tildeMax(y12, mu1, mu2, mu3);
    double y23StepWidth = (y23max - y23min)/numberofsteps;
    // j=0 and j=numberofsteps are on the boundary of the phase space,
    // omit them
    for (int j=1; j<=numberofsteps-1; j++) {
      double y23 = y23min + j*y23StepWidth;
      // Evaluate a2 with and without finite terms, enable
      // a2 to return negative values
      double aFull   = a2(y12,y23, mu1, mu2, mu3, false);
      bool useFinSav = useFiniteTerms;
      useFiniteTerms = false;
      double aSing   = a2(y12,y23, mu1, mu2, mu3, false);
      // reset the use of the finite terms
      useFiniteTerms = useFinSav;
      if ( aFull <= - TINY ) {
	isPos = false;
	if (dorepair) {
	  // if the singular part of the antenna is negative 
	  // and the function is supposed to repair the 
	  // antenna, there is a problem
	  if (aSing <= -TINY) {
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 1) {
	      printOut(name+"a2:IsPositive",
		       string("Singular part of antenna ")
		       + " negative.\n"
		       + "\ty12 = " + num2str(y12, 4) +"\n"
		       + "\ty23 = " + num2str(y12, 4) +"\n"
		       + "\tmu1 = " + num2str(mu1, 4) +"\n"
		       + "\tmu2 = " + num2str(mu2, 4) +"\n"
		       + "\tmu3 = " + num2str(mu3, 4) +"\n"
		       + "\taFull = " + num2str(aFull, 4) +"\n"
		       + "\taSing = " + num2str(aSing, 4) +"\n"
		       );
	    }
	    verbose = verboseSav;
	    return false;
	  }
	  else {
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 5) {
	      printOut(name+":IsPositive",
		       "rescaling finite terms to make "
		       + string("antenna positive")
		       );
	    }
	    bool rescaleSucessful =
	      this->rescaleFiniteTerms(0.9);
	    if (rescaleSucessful) {
	      // restart the check with the changed 
	      // finite terms
	      if (verboseSav >= 3) {
		printOut(name+":IsPositive",
			 "new coefficients are:");
		this->listCoefficients();
	      }
	      verbose = verboseSav;
	      return this->IsPositive(nPointsCheck, mu1, 
				      mu2, mu3, true);
	    }
	    else {
	      // if it has not been possible to rescale
	      // the finite terms, complain and return
	      // false
	      // verbose has been set to -1, 
	      // use verboseSav
	      if (verboseSav >= 1) {
		printOut(name+"a2:IsPositive",
			 "Could not rescale finite "
			 + string("terms.\n") );
	      }
	      verbose = verboseSav;
	      return false;
	    }
	  }
	}
	else {
	  // verbose has been set to -1, use verboseSav
	  if (verboseSav >= 3) {
	    cout << "IsPositive::negative value" << endl;
	    cout << "Antenna name = " << getHumanName();
	    cout << endl;
	    cout << scientific;
	    cout << setprecision(20);
	    cout << "mu1 = " << mu1 << endl;
	    cout << "mu2 = " << mu2 << endl;
	    cout << "mu3 = " << mu3 << endl;
	    cout << "y12 = " << y12 << endl;
	    cout << "y23 = " << y23 << endl;
	    cout << "full antenna = " << aFull << endl;
	    cout << "singular part = " << aSing << endl;
	  }
	}
      }
    } // end of for (int j=0; j<=numberofsteps...
  }
  // reset the verbosity
  verbose = verboseSav;
  return isPos;
}

// MR 21.12.2010: IsFreeOfDeadZones checks whether an antenna has dead
// zones given a set of masses. It has an optional argument dorepair 
// (default false) which governs whether the finite coefficients are 
// changed to make the dead zones go away.
bool Antenna::IsFreeOfDeadZones( int nPointsCheck, double mu1, 
				 double mu2, double mu3, 
				 bool dorepair ) {
  // initialize the flags
  bool isFreeOfDeadZones = true;
  
  // don't want the antenna to complain, set verbosity to -1
  int verboseSav = verbose;
  verbose = -1;
  
  // calculate the phase space boundaries
  double y12min = 2*mu1*mu2;
  double y12max = pow2(1 - mu3) - pow2(mu1) - pow2(mu2);
  // loop through the phase space
  int numberofsteps = floor( sqrt( nPointsCheck ) );
  double y12StepWidth = (y12max - y12min)/numberofsteps;
  // i=0 and i=numberofsteps are the boundary of the phase space,
  // omit them
  for (int i=1; i<=numberofsteps-1; i++) {
    double y12 = y12min + i*y12StepWidth;
    // make sure y12 is positive
    y12 = max(y12,0.);
    double y23min = y23tildeMin(y12, mu1, mu2, mu3);
    double y23max = y23tildeMax(y12, mu1, mu2, mu3);
    double y23StepWidth = (y23max - y23min)/numberofsteps;
    // j=0 and j=numberofsteps are the boundary of the phase space,
    // omit them
    for (int j=1; j<=numberofsteps-1; j++) {
      double y23 = y23min + j*y23StepWidth;
      // make sure y23 is positive
      y23 = max(y23, 0.);
      // Evaluate a2 with and without finite terms, enable
      // a2 to return negative values
      double aFull   = a2(y12,y23, mu1, mu2, mu3, false);
      bool useFinSav = useFiniteTerms;
      useFiniteTerms = false;
      double aSing   = a2(y12,y23, mu1, mu2, mu3, false);
      // reset the use of the finite terms
      useFiniteTerms = useFinSav;
      // only do something if aFull< deadZoneAvoidance*aSing
      if ( aFull < deadZoneAvoidance*aSing ) {
	isFreeOfDeadZones = false;
	if (dorepair) {
	  // rescale
	  if ( this->rescaleFiniteTerms(0.9) ) {
	    // restart the check with the changed 
	    // finite terms
	    verbose = verboseSav;
	    return this->IsFreeOfDeadZones(nPointsCheck, 
					   mu1, mu2, mu3, true);
	  }
	  else {
	    // if it has not been possible to rescale
	    // the finite terms, complain and return
	    // false
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 1) {
	      printOut(name+"a2:IsFreeOfDeadZones",
		       "Could not rescale finite terms.\n");
	    }
	    verbose = verboseSav;
	    return false;
	  }
	}
	else {
	  // verbose has been set to -1, use verboseSav
	  if (verboseSav >= 3) {
	    cout << "IsFreeOfDeadZones::dead zone" << endl;
	    cout << "Antenna name = " << getHumanName();
	    cout << endl;
	    cout << scientific;
	    cout << setprecision(20);
	    cout << "mu1 = " << mu1 << endl;
	    cout << "mu2 = " << mu2 << endl;
	    cout << "mu3 = " << mu3 << endl;
	    cout << "y12 = " << y12 << endl;
	    cout << "y23 = " << y23 << endl;
	    cout << "full antenna = " << aFull << endl;
	    cout << "singular part = " << aSing << endl;
	    cout << "dead Zone avoidance factor = " 
		 << deadZoneAvoidance << endl;
	  }
	}
      }
    }
  }
  // reset the verbosity
  verbose = verboseSav;
  return isFreeOfDeadZones;
}

// MR 21.12.2010: rescaleFiniteTerms rescales all the finite terms
// of an antenna, returning true if it was able to do so, 
// false otherwise
  bool Antenna::rescaleFiniteTerms(double rescalingFactor) {
    if (inputMode == 1) {
      // list the old parameter values
      if (verbose >= 5) {
	printOut(name+":rescaleFiniteTerms", 
		 string("reducing magnitude of finite terms.\n")
		 + "old parameter values:\n"
		 );
	listCoefficients();
      }
      // scale C(i,j) terms
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
	int ic=it->first;
	for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
	  int jc=jt->first;  
	  if ( ic >= 0 && jc >= 0) {
	    double oldcoeff = jt->second;
	    double newabs = rescalingFactor*abs(oldcoeff);
	    // round the new absolute value down to the closest
	    // multiple of abs(1-rescalingFactor)
	    double minchange = abs(1-rescalingFactor);
	    newabs = floor( newabs/minchange )*minchange;
	    setC(ic, jc,oldcoeff*newabs/abs(oldcoeff) );
	  }
	}
      } // end of loop through C(i,j) terms
	// scale down the MNum terms
      for (map< SixInt, double>::iterator pt=LaurentMNum.begin();
	   pt!=LaurentMNum.end();++pt) {
	int y12ord = pt->first.int1;
	int y13ord = pt->first.int2;
	int y23ord = pt->first.int3;
	int mu1ord = pt->first.int4;
	int mu2ord = pt->first.int5;
	int mu3ord = pt->first.int6;
	double oldcoeff = pt->second;
	double newabs = rescalingFactor*abs(oldcoeff);
	// round the new absolute value down to the closest
	// multiple of abs(1-rescalingFactor)
	double minchange = abs(1-rescalingFactor);
	newabs = floor( newabs/minchange )*minchange;
	setMNum(y12ord, y13ord, y23ord, mu1ord, mu2ord, mu3ord,
		oldcoeff*newabs/abs( oldcoeff ) );
	
      }
      // no other terms included in scaling down at the moment
      return true;
    } // end of if (inputMode == 1)
    else {
      if (verbose >= 3) {
	printOut(name+":rescaleFiniteTerms", 
		 "inputMode " + num2str( inputMode )
		 + " not recognized.");
      }
      return false;
    }
  }


// MR 16.7.2010: for debugging purposes, have the antenna list
// its coefficients on an output stream
void Antenna::listCoefficients(ostream& os) {
  os << setprecision(3);
  os << "antenna " << getHumanName() << "\n";
  
  os << "use finite Terms: ";
  if (useFiniteTerms) {
    os << "True" << endl;
  }
  else {
    os << "False" << endl;
  }
  
  map<int,map<int,double> >::iterator it;
  map<int,double>::iterator jt;
  // C(i,j) terms
  for (it=LaurentC.begin();it!=LaurentC.end();++it) {
    int i=it->first;
    for (jt=LaurentC[i].begin();jt!=LaurentC[i].end();++jt) {
      int j=jt->first;  
      double c=jt->second;
      os << "LaurentC(" << i << "," << j << ") = " << c << "\n";   
    }      
  }
  // D(j) terms
  for (jt=LaurentD.begin();jt!=LaurentD.end();++jt) {
    int j=jt->first;
    double d=jt->second;
    os << "LaurentD(" << j << ") = " << d << "\n";
  }
  // E(j) terms
  for (jt=LaurentE.begin();jt!=LaurentE.end();++jt) {
    int j=jt->first;
    double e=jt->second;
    os << "LaurentE(" << j << ") = " << e << "\n";
  }
  // K term
  if (LaurentK != 0) {
    os << "LaurentK = " << LaurentK << "\n";
  }
  // CollinearGluonSplitting term
  if (CollinearGluonSplitting != 0) {
    os << "CollinearGluonSplitting = " << CollinearGluonSplitting << "\n";
  }
  
  
  // MNum coefficients
  for (map< SixInt, double>::iterator pt=LaurentMNum.begin(); pt!=LaurentMNum.end();++pt) {
    int y12ord = pt->first.int1;
    int y13ord = pt->first.int2;
    int y23ord = pt->first.int3;
    int mu1ord = pt->first.int4;
    int mu2ord = pt->first.int5;
    int mu3ord = pt->first.int6;
    double presentcoeff = pt->second;
    os << "LaurentMNum(" << y12ord << "," << y13ord << ","
       << y23ord << "," << mu1ord << "," << mu2ord << ","
       << mu3ord << ") = " << presentcoeff << "\n";
  }
  
  
}
  
} // End Vincia Namespace
  
