// VinciaEvolutionVariables.cc is a part of the VINCIA plug-in to the 
// PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the EvolutionVariable class

// Include the VinciaEvolutionVariables header
#include "VinciaEvolutionVariables.h"

using namespace Pythia8;

namespace Vincia {
  
// Definition of the dimensionless resolution function

double EvolutionVariable::yE(double y12, double y23, int typeIn) {

  // Default type == evolution variable
  if (typeIn == 0) typeIn = settingsPtr->mode("VinciaShower:evolutionType");

  double yENow = 1.;

  // Type 1 : pT-ordering (ARIADNE)
  if (typeIn == 1) {
    yENow= 4.0*y12*y23;
  } 

  // Type 2 : daughter mass-ordering (mAnt-ordering)
  else if (typeIn == 2) {
    yENow= 2.0*min(y12,y23);
  } 

  // Type 3 : energy ordering 
  // Warning: not infrared safe a priori, but combined with
  // hadronization cut in pT can still be used.
  else if (typeIn == 3) {
    yENow= pow2(y12 + y23);
  } 

  // Type 4 : anticollinear (Upsilon) ordering
  else if (typeIn == 4) {
    yENow= pow(pow2(y12+y23),0.25)-pow(pow2(y12-y23),0.25);
  }

  // Type 5 : collinear-safe energy ordering 
  else if (typeIn == 5) {
    yENow= sqrt(pow2(pow2(y12+y23))-pow2(pow2(y12-y23)));
  }

  // Type 6 and -1 : Pythia pTevol = min(pTevolA,pTevolB)
  else if (typeIn == -1 || typeIn == 6) {
    double yEvolA = 4.*y12*(1-y23)*(y12+y23)/pow2(1-y12); 
    double yEvolB = 4.*y23*(1-y12)*(y23+y12)/pow2(1-y23); 
    yENow= min(yEvolA,yEvolB);
  }

  
  //// sector shower
  // Type 8: sector shower for quarks with color-ordered clustering (gluon /phys. quark , quark, quark companion)
  else if (typeIn == 8) {
    yENow= sqrt(y12)*y23;
  }
  // Type 9: sector shower for quarks with color-ordered clustering (quark companion, quark, gluon /phys. quark)
  else if (typeIn == 9) {
    yENow= y12*sqrt(y23);
  }
  // Type 10: sector shower for gluons
  else if (typeIn == 10) {
    yENow= y12*y23;
  }
  //////////////
  
  
  // Type 11 : phase-space ordering / virtuality-ordering 
  // Warning: not infrared safe
  else if (typeIn == 11) {
    double yMax = max(y12,y23);
    yENow= yMax;
  } 

  // Type 12 : experimental Rapidity-like ordering (warning: not infrared safe)
  // Divide phase space into y12 > y23 and y12 < y23
  else if (typeIn == 12) {
    double yRap=0.0;
    if (y12 < y23) {
      double ct23 = 1-y23/2.0/(1-y12)/(y12+y23);
      double trap = M_PI - acos(ct23);
      yRap = sqrt((1.0-cos(trap))/(1.0+cos(trap)));      
    } else {
      double ct12 = 1-y12/2.0/(1-y23)/(y12+y23);
      double trap = M_PI - acos(ct12);
      yRap = sqrt((1.0-cos(trap))/(1.0+cos(trap)));      
    }
    yENow= yRap;
  } 

  // Type 13 : Thrust-ordering. 
  // Warning: not infrared safe
  else if (typeIn == 13) {

    yENow= 3.0*min(min(y12,y23),1.-y12-y23);

  } 

  // Type 15 : experimental angular-like ordering (warning: not infrared safe)
  // Divide phase space into y12 > y23 and y12 < y23
  else if (typeIn == 15) {
    if (y12 < y23) {
      yENow= max(y12/(1-y23),y12/y23);      
    } else {
      yENow= max(y23/(1-y12),y23/y12);      
    }
  } 

  // Print warning if type not recognized at all
  else {
    if (verbose >= 1)
      printErr("Vincia::EvolutionVariable::R",
	       "Unknown ordering variable :"+num2str(typeIn));
    yENow= -1.;
  }

  // Return
  return yENow;

}
  
  // Copy of R using 3 Vec4 as input 
  double EvolutionVariable::yE(Vec4 v1, Vec4 v2, Vec4 v3, int typeIn) {
    
    // Compute invariants (including masses)
    // MR 12.8.2010: changed to sij=2*pi*pj and s = (p1+p2+p3)^2
    double s=m2(v1,v2,v3);
    if (s == 0.0) return 0.0;
    double y12=(2*v1*v2)/s;
    double y23=(2*v2*v3)/s;
    return yE(y12,y23,typeIn);    
    
  }

  // Copy of R using 3 double[4] as input 
  double EvolutionVariable::yE(double* p1,double* p2,double* p3, int typeIn) {
    
    // Translate from doubles to Vec4, then call Vec4 function.
    Vec4 v1(p1[0],p1[1],p1[2],p1[3]),v2(p2[0],p2[1],p2[2],p2[3]),
      v3(p3[0],p3[1],p3[2],p3[3]);
    return yE(v1,v2,v3,typeIn);
    
  }

  // Copy for returning R*s
  double EvolutionVariable::Q2E(Vec4 v1, Vec4 v2, Vec4 v3, int typeIn) { 
    // MR 24.2.2011: implement a "massive p_T" on code 42 for testing
    if (typeIn == 42) {
      double s=m2(v1,v2,v3);
      double y12=m2(v1, v2)/s;
      double y23=m2(v2, v3)/s;
      return yE(y12,y23,1)*s;
    }
    
    // Compute invariants (including masses)
    // MR 12.8.2010: changed to sij=2*pi*pj and s = (p1+p2+p3)^2
    double s=m2(v1,v2,v3);
    if (s == 0.0) return 0.0;
    double y12=(2*v1*v2)/s;
    double y23=(2*v2*v3)/s;
    return yE(y12,y23,typeIn)*s;
    
  }
  
  // Copy for returning R*s from Particles
  double EvolutionVariable::Q2E(Particle p1, Particle p2, Particle p3, int typeIn) {
    
    return Q2E(p1.p(),p2.p(),p3.p(),typeIn);
    
  }

  // Copy for returning R*s from y12, y23, and s
  double EvolutionVariable::Q2E(double s, double y12, double y23, int typeIn) {
    
    return s*yE(y12,y23,typeIn);

  }

}
