// vincia04-root is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study n-jet rates and the so-called
// 4-jet angles LEP 1, using VINCIA's ROOT-based runtime display. 
// Uses VINCIA for showering and matching, FASTJET for jet clustering, 
// and PYTHIA 8 for all else
// Copyright (c) 2011 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot.h"
// FASTJET
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythia; 
  VinciaPlugin vincia(&pythia,"vincia04.cmnd");
  // Uncomment the next line instead to run standalone Pythia for comparison:
  // VinciaPlugin vincia(&pythia,"pythia01-root.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event& event = pythia.event;

  // Some useful pointers
  Settings* settingsPtr = vincia.settingsPtr;        // to settings database
  //VinClu* vinCluPtr = &vincia.vinClu;                // to Vincia's 3->2 clu
  //Matching* matchingPtr = &vincia.matching;          // to Vincia's matching
  //MGInterface* mgInterfacePtr = &vincia.mgInterface; // to the MG interface

  // Extract settings to be used in the main program.  
  double eCM       = settingsPtr->parm("Beams:eCM");
  int    nEvent    = settingsPtr->mode("Main:numberOfEvents");
  int    nShow     = settingsPtr->mode("Main:timesToShow");

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");
  string antennaFile      = settingsPtr->word("VinciaShower:antennaFile");
  bool   uncertaintyBands = vinciaOn && settingsPtr->flag("Vincia:uncertaintyBands");
  int    verbose          = settingsPtr->mode("Vincia:verbose");
  int    orderingMode     = settingsPtr->mode("VinciaShower:evolutionMode");
  int    showerNLC        = settingsPtr->mode("VinciaShower:NLC");

  // Extrac matching settings
  int    maxMatchTree     = settingsPtr->mode("VinciaMatching:order");
  bool   matching         = maxMatchTree >= 1;
  bool   scaleMatching    = settingsPtr->mode("VinciaShower:NLL") >= 1;
  int    matchingRegOrder = settingsPtr->mode("VinciaMatching:regOrder");

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create VinciaRoot object 
  VinciaRoot vroot; 
  // Initialize some ROOT displays (name,x-pixels,y-pixels,x-offset,y-offset)
  // Book additional displays here as you like.
  vroot.createDisplay("Resolutions",620,780);
  vroot.createDisplay("Rates",620,780);
  vroot.createDisplay("Angles",600,750);
  vroot.createDisplay("Fragmentation",300,400);

  // Tell VinciaRoot which verbose level to use
  vroot.setVerbose(verbose);

 // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose(verbose);
  // Switch stats on/off (in ROOT conventions, but with 0 = off)
  vroot.setOptStat(0);

  // Tell it the name of the MC we are running (for plot legends)
  string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),4);  
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),4);  
  if (vinciaOn && matching) {
    vroot.setLabel("Vincia");
    vroot.setRefs("Vincia 1.025 + MadGraph 4.426 + Pythia "+pyVer);
  } else if (vinciaOn) {
    vroot.setLabel("Vincia");
    vroot.setRefs("Vincia 1.025 + Pythia "+pyVer);
  } else {
    vroot.setLabel("Pythia 8");
    vroot.setRefs("Pythia "+pyVer);
  }
  // Set names for uncertainty bands, if using them (for plot legends)
  if (uncertaintyBands) {
    vroot.addUncertainty("Def",0,1);
    vroot.addUncertainty("#muR",2,3);
    vroot.addUncertainty("Finite",4,5);
    vroot.addUncertainty("QMatch",6,7);
    vroot.addUncertainty("Ord",8,9);
    vroot.addUncertainty("1/N_{C}^{2}",10,11);
  }

  // Use a different color scheme for PYTHIA plots
  if ( !vinciaOn ) vroot.setColorScheme(2);
  // Use a different color scheme if uncertainty bands off
  else if ( !uncertaintyBands ) vroot.setColorScheme(3);

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  pythia.init( 11, -11, eCM);

  // List changed settings
  pythia.settings.listChanged();

  // Check that Z0 decay channels set correctly.
  pythia.particleData.listChanged();
 
  //***************************************************************************
  // FASTJET
  //***************************************************************************

  // Fastjet analysis - select algorithm and parameters
  fastjet::Strategy               strategy = fastjet::Best;
  fastjet::RecombinationScheme    recombScheme = fastjet::E_scheme;
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::ee_kt_algorithm,
                                      recombScheme, strategy);
  // Fastjet input
  std::vector <fastjet::PseudoJet> fjInputs;

  //***************************************************************************
  // ROOT
  //***************************************************************************

  // Form name for ROOT file
  ostringstream outfile;
  // Put root files in subdirectory defined at beginning of file
  if (not vinciaOn) {
    outfile<<"pythia04";
  } else {
    outfile<<"vincia04";
  }
  if (settingsPtr->flag("HadronLevel:all")) outfile<<"-had";
  outfile<<".root";

  //****************
  // Book Histograms

  // Vincia uncertainty bands: number of variations expected 
  int nVariations = max(1,vincia.nVariations());

  // Jet Rates
  int nBins  = 6;
  double xLo = -1.2;
  double xHi =  0.0;
  vector<TH1F*> Vr1j;
  Vr1j.push_back(new TH1F("r1j","Exclusive 1-Jet Rate",nBins, xLo, xHi));
  // Add to display with
  // logy = true, normalize=false, showStatErr=true
  Vr1j[0]->SetMaximum(1.01); 
  Vr1j[0]->SetMinimum(0.00004); 
  Vr1j[0]->GetXaxis()->SetTitle("Ln(y1)");
  Vr1j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr1j.push_back(new TH1F(string("r1j_"+num2str(iVar,1)).c_str()," ",
			    nBins,xLo,xHi));  
  vroot.addTH1FPtr(Vr1j,"Rates",true,false,true,"data/Aleph-91-r1j.vec","pythia-r1j.dat");

  // Jet Rates
  nBins = 51;
  xLo   = -10.2;
  xHi   =   0.0;
  vector<TH1F*> Vr2j;
  Vr2j.push_back(new TH1F("r2j","Exclusive 2-Jet Rate",nBins,xLo,xHi));
  Vr2j[0]->SetMaximum(200.01); 
  Vr2j[0]->SetMinimum(0.00004); 
  Vr2j[0]->GetXaxis()->SetTitle("Ln(y2)");
  Vr2j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr2j.push_back(new TH1F(string("r2j_"+num2str(iVar,1)).c_str()," ",
				    nBins,xLo,xHi));  
  vroot.addTH1FPtr(Vr2j,"Rates",true,false,true,"data/Aleph-91-r2j.vec","pythia-r2j.dat");

  // Jet Rates
  nBins =    46;
  xLo   = -10.2;
  xHi   =  -1.0;
  vector<TH1F*> Vr3j;
  Vr3j.push_back(new TH1F("r3j","Exclusive 3-Jet Rate",nBins,xLo,xHi));
  Vr3j[0]->SetMaximum(200.01); 
  Vr3j[0]->SetMinimum(0.00002); 
  Vr3j[0]->GetXaxis()->SetTitle("Ln(y3)");
  Vr3j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr3j.push_back(new TH1F(string("r3j_"+num2str(iVar,1)).c_str()," ",
				    nBins,xLo,xHi));  
  vroot.addTH1FPtr(Vr3j,"Rates",true,false,true,"data/Aleph-91-r3j.vec","pythia-r3j.dat");

  // Jet Rates
  nBins =    41;
  xLo   = -10.2;
  xHi   =  -2.0;
  vector<TH1F*> Vr4j;
  Vr4j.push_back(new TH1F("r4j","Exclusive 4-Jet Rate",nBins,xLo,xHi));
  Vr4j[0]->SetMaximum(120.01); 
  Vr4j[0]->SetMinimum(0.000005); 
  Vr4j[0]->GetXaxis()->SetTitle("Ln(y4)");
  Vr4j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr4j.push_back(new TH1F(string("r4j_"+num2str(iVar,1)).c_str()," ",
				    nBins,xLo,xHi));  
  vroot.addTH1FPtr(Vr4j,"Rates",true,false,true,"data/Aleph-91-r4j.vec","pythia-r4j.dat");

  // Jet Rates
  nBins =    36;
  xLo   = -10.2;
  xHi   =  -3.0;
  vector<TH1F*> Vr5j;
  Vr5j.push_back(new TH1F("r5j","Exclusive 5-Jet Rate",nBins,xLo,xHi));
  Vr5j[0]->SetMaximum(20.01); 
  Vr5j[0]->SetMinimum(0.000005); 
  Vr5j[0]->GetXaxis()->SetTitle("Ln(y5)");
  Vr5j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr5j.push_back(new TH1F(string("r5j_"+num2str(iVar,1)).c_str()," ",
				    nBins,xLo,xHi));  
  vroot.addTH1FPtr(Vr5j,"Rates",true,false,true,"data/Aleph-91-r5j.vec","pythia-r5j.dat");

  // Jet Rates
  nBins =    31;
  xLo   = -10.2;
  xHi   =  -4.0;
  vector<TH1F*> Vr6j;
  Vr6j.push_back(new TH1F("r6j","Inclusive 6-Jet Rate",nBins,xLo,xHi));
  Vr6j[0]->SetMaximum(20.01); 
  Vr6j[0]->SetMinimum(0.000005); 
  Vr6j[0]->GetXaxis()->SetTitle("Ln(y6)");
  Vr6j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr6j.push_back(new TH1F(string("r6j_"+num2str(iVar,1)).c_str()," ",
				    nBins,xLo,xHi));  
  vroot.addTH1FPtr(Vr6j,"Rates",true,false,true,"data/Aleph-91-r6j.vec","pythia-r6j.dat");

  // Jet Resolutions
  nBins = 13;
  xLo   = 0.0;
  xHi   = 1.3;
  double wY12 = nBins/(xHi-xLo);
  vector<TH1F*> VLnY12;
  VLnY12.push_back(new TH1F("LnY12","Ln(1/y12)",nBins, xLo, xHi));
  VLnY12[0]->SetMaximum(20.1); 
  VLnY12[0]->SetMinimum(0.000005); 
  VLnY12[0]->GetXaxis()->SetTitle("Ln(1/y12)");
  VLnY12[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VLnY12.push_back(new TH1F(string("LnY12_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VLnY12,"Resolutions",true,true,true,"data/LEP91-LnY12.vec","pythia-lny12.dat");

  nBins = 51;
  xLo = 0.8;
  xHi = 11.0;
  double wY23 = nBins/(xHi-xLo);
  vector<TH1F*> VLnY23;
  VLnY23.push_back(new TH1F("LnY23","Ln(1/y23)",nBins, xLo, xHi));
  VLnY23[0]->SetMaximum(20.01); 
  VLnY23[0]->SetMinimum(0.00005); 
  VLnY23[0]->GetXaxis()->SetTitle("Ln(1/y23)");
  VLnY23[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VLnY23.push_back(new TH1F(string("LnY23_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VLnY23,"Resolutions",true,true,true,"data/LEP91-LnY23.vec","pythia-lny23.dat");

  nBins = 48;
  xLo = 1.8;
  xHi = 11.4;
  double wY34 = nBins/(xHi-xLo);
  vector<TH1F*> VLnY34;
  VLnY34.push_back(new TH1F("LnY34","Ln(1/y34)",nBins, xLo, xHi));
  VLnY34[0]->SetMaximum(20.01); 
  VLnY34[0]->SetMinimum(0.00005); 
  VLnY34[0]->GetXaxis()->SetTitle("Ln(1/y34)");
  VLnY34[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VLnY34.push_back(new TH1F(string("LnY34_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VLnY34,"Resolutions",true,true,true,"data/LEP91-LnY34.vec","pythia-lny34.dat");

  nBins = 45;
  xLo = 3.0;
  xHi = 12.0;
  double wY45 = 200.0/7.0;
  vector<TH1F*> VLnY45;
  VLnY45.push_back(new TH1F("LnY45","Ln(1/y45)",nBins, xLo, xHi));
  VLnY45[0]->SetMaximum(20.01); 
  VLnY45[0]->SetMinimum(0.00005); 
  VLnY45[0]->GetXaxis()->SetTitle("Ln(1/y45)");
  VLnY45[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VLnY45.push_back(new TH1F(string("LnY45_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VLnY45,"Resolutions",true,true,true,"data/LEP91-LnY45.vec","pythia-lny45.dat");

  nBins = 46;
  xLo = 3.8;
  xHi = 13.0;
  double wY56 = 200.0/6.;
  vector<TH1F*> VLnY56;
  VLnY56.push_back(new TH1F("LnY56","Ln(1/y56)",nBins, xLo, xHi));
  VLnY56[0]->SetMaximum(20.01); 
  VLnY56[0]->SetMinimum(0.00005); 
  VLnY56[0]->GetXaxis()->SetTitle("Ln(1/y56)");
  VLnY56[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VLnY56.push_back(new TH1F(string("LnY56_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VLnY56,"Resolutions",true,true,true,"data/LEP91-LnY56.vec","pythia-lny56.dat");

  // BZ Angle
  nBins = 20;
  xLo   = 0.0;
  xHi   = 1.0;
  double wBZ = 20.0/1.0;
  vector<TH1F*> VChiBZ;
  VChiBZ.push_back(new TH1F("CosChiBZ","Bengtsson-Zerwas Angle",nBins, xLo, xHi));
  VChiBZ[0]->SetMaximum(4.0); 
  VChiBZ[0]->SetMinimum(-0.2); 
  VChiBZ[0]->GetXaxis()->SetTitle("cos(#Chi_{BZ})");
  VChiBZ[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VChiBZ.push_back(new TH1F(string("ChiBZ_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VChiBZ,"Angles",false,true,true,"data/Delphi-91-chiBZ.vec","pythia-chiBZ.dat");

  // KSW Angle
  nBins = 20;
  xLo   = -1.0;
  xHi   =  1.0;
  double wKSW = 20.0/2.0;
  vector<TH1F*> VPhiKSW;
  VPhiKSW.push_back(new TH1F("CosPhiKSW","Korner-Schierholz-Willrodtorner Angle",
			     nBins, xLo, xHi));
  VPhiKSW[0]->SetMaximum(1.5); 
  VPhiKSW[0]->SetMinimum(-0.075); 
  VPhiKSW[0]->GetXaxis()->SetTitle("cos(#Phi_{KSW})");
  VPhiKSW[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VPhiKSW.push_back(new TH1F(string("PhiKSW_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VPhiKSW,"Angles",false,true,true,"data/Delphi-91-phiKSW.vec","pythia-phiKSW.dat");

  // NR Angle
  nBins = 20;
  xLo   =  0.0;
  xHi   =  1.0;
  double wNR = 20.0/1.0;
  vector<TH1F*> VThetaNR;
  VThetaNR.push_back(new TH1F("CosThetaNR","Nachtman-Reiter Angle", nBins, xLo, xHi));
  VThetaNR[0]->SetMaximum(2.5); 
  VThetaNR[0]->SetMinimum(-0.125); 
  VThetaNR[0]->GetXaxis()->SetTitle("|cos(#Theta_{NR})|");
  VThetaNR[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VThetaNR.push_back(new TH1F(string("ThetaNR_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VThetaNR,"Angles",false,true,true,"data/Delphi-91-thetaNR.vec","pythia-thetaNR.dat");

  // 34 Angle
  nBins = 20;
  xLo   = -1.0;
  xHi   = 1.0;
  double wA34 = 20.0/2.0;  
  vector<TH1F*> VAlpha34;
  VAlpha34.push_back(new TH1F("CosAlpha34","#alpha_{34}", nBins, xLo, xHi));
  VAlpha34[0]->SetMaximum(1.2); 
  VAlpha34[0]->SetMinimum(-0.06); 
  VAlpha34[0]->GetXaxis()->SetTitle("cos(#alpha_{34})");
  VAlpha34[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
		  VAlpha34.push_back(new TH1F(string("Alpha34_"+num2str(iVar,1)).c_str()," ",
					  nBins,xLo,xHi));  
  vroot.addTH1FPtr(VAlpha34,"Angles",false,true,true,"data/Delphi-91-a34.vec","pythia-alpha34.dat");

  // nCharged
  TH1F *HnCharged=new TH1F("nCh","nCharged", 60, -0.5, 59.5);
  //vroot.addTH1FPtr(HnCharged,"Fragmentation",true,true,true,"data/Aleph-91-Nch.vec");
  HnCharged->SetMinimum(0.00001);
  HnCharged->SetMaximum(0.5);

  // Ln(1/x)
  TH1F *Hx=new TH1F("LnXInv","Ln(1/x)", 200, 0.0, 5.4);
  double wx = 200.0/5.4;
  vroot.addTH1FPtr(Hx,"Fragmentation",true,true,true,"data/Aleph-91-LogInvx.vec");
  Hx->SetMinimum(0.0002);
  Hx->SetMaximum(8.);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settingsPtr->mode("Random:seed");
  vincia.rndmPtr->init(initseed);

  // Counter for negative-weight events
  double nNegative = 0.0;
  double nPositive = 0.0;
  double evwt=1.0;
  double evwtMax=1.0;
  double evwtMin=1.0;
  double sumw[nVariations];
  for (int iVar=0; iVar<nVariations; iVar++) sumw[iVar]=0.0;

  // Counters for jet rates
  nBins = 100;
  double sumN1j[nBins][nVariations];  
  double sumN2j[nBins][nVariations];
  double sumN3j[nBins][nVariations];
  double sumN4j[nBins][nVariations];
  double sumN5j[nBins][nVariations];
  double sumN6j[nBins][nVariations];
  for (int iBin=0; iBin<nBins; iBin++) 
    for (int iVar=0; iVar<nVariations; iVar++) {
      sumN1j[iBin][iVar] = 0.0;
      sumN2j[iBin][iVar] = 0.0;
      sumN3j[iBin][iVar] = 0.0;
      sumN4j[iBin][iVar] = 0.0;
      sumN5j[iBin][iVar] = 0.0;
      sumN6j[iBin][iVar] = 0.0;
    }


  // FastJet needs bool to tell whether this event is the first one
  bool firstEvent = true; 

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {


    if (iEvent%(max(1,nEvent/max(nShow,1))) == 0)  
      cout << " Now begin event " << iEvent << endl;

    if (!pythia.next()) continue;

    // Check for negative weights
    if (vinciaOn) {
      evwt    = vincia.weight();
      evwtMax = vincia.weight(2);
      evwtMin = vincia.weight(3);
      //      cout << evwt <<" "<<evwtMax<<" "<<evwtMin<<endl;
    }
    if (evwt < 0) nNegative += abs(evwt); 
    else nPositive += evwt;
    for (int iVar=0; iVar<nVariations; iVar++) sumw[iVar] += vincia.weight(iVar);

    // List the first few events
    if (iEvent < 3) event.list();

    // Count FS charged hadrons, partons, and quarks
    int nc=0;

    // Add final-state particles to FastJet input vector    
    fjInputs.resize(0);
    for (int i=1;i<event.size();i++) {

      // Final state only
      if (!pythia.event[i].isFinal())        continue;
      
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) {
	nc++;
	double LnXInv=log(double(1.0/(2.0*event[i].pAbs()/eCM)));
	Hx->Fill(LnXInv,evwt*wx);
      }

      // Now decide what to pass to FastJet
      // NOTE: neutrinos are *included* in this list, since ALEPH
      // appears to have corrected their data to include them.
      // Store as input to Fastjet
      fjInputs.push_back( fastjet::PseudoJet (pythia.event[i].px(),
					      pythia.event[i].py(), 
					      pythia.event[i].pz(),
					      pythia.event[i].e() ) );
    }

    // FastJet
    if (fjInputs.size() == 0) {
      cout << "Error: event with no final state particles" << endl;
      continue;
    }

    // Run Fastjet algorithm    
    fastjet::ClusterSequence clustSeq(fjInputs, *jetDef);

    // Use yCut from LEP analysis
    vector<fastjet::PseudoJet> fastjets =  clustSeq.exclusive_jets_ycut(0.008);
    vector<fastjet::PseudoJet> sorted = fastjet::sorted_by_E(fastjets);

    // Extract merging scales
    double lny56 = log(clustSeq.exclusive_ymerge(5));
    double lny45 = log(clustSeq.exclusive_ymerge(4));
    double lny34 = log(clustSeq.exclusive_ymerge(3));
    double lny23 = log(clustSeq.exclusive_ymerge(2));
    double lny12 = log(clustSeq.exclusive_ymerge(1));

    // Fill Jet resolution histograms with results from FastJet
    for (int iVar=0; iVar < nVariations; iVar++) {
      VLnY12[iVar]->Fill(-lny12,wY12*vincia.weight(iVar));
      VLnY23[iVar]->Fill(-lny23,wY23*vincia.weight(iVar));
      VLnY34[iVar]->Fill(-lny34,wY34*vincia.weight(iVar));
      VLnY45[iVar]->Fill(-lny45,wY45*vincia.weight(iVar));
      VLnY56[iVar]->Fill(-lny56,wY56*vincia.weight(iVar));
    }

    // Histogram jet rates
    // Loop over resolutions
    for (int iVar=0; iVar < nVariations; iVar++) {
      Vr1j[iVar]->Reset();
      Vr2j[iVar]->Reset();
      Vr3j[iVar]->Reset();
      Vr4j[iVar]->Reset();
      Vr5j[iVar]->Reset();
      Vr6j[iVar]->Reset();
      int iy=-1;
      for (double lny=-10.1; lny<0.0; lny += 0.2) {
	iy++;
	if (lny12 < lny) sumN1j[iy][iVar] += vincia.weight(iVar);
	else if (lny23 < lny) sumN2j[iy][iVar] += vincia.weight(iVar);
	else if (lny34 < lny) sumN3j[iy][iVar] += vincia.weight(iVar);
	else if (lny45 < lny) sumN4j[iy][iVar] += vincia.weight(iVar);
	else if (lny56 < lny) sumN5j[iy][iVar] += vincia.weight(iVar);
	else  sumN6j[iy][iVar] += vincia.weight(iVar);
	double err, val;
	// 1j
	val = sumN1j[iy][iVar]/sumw[iVar];
	err = sqrt(sumN1j[iy][iVar])/sumw[iVar];
	int iBin = Vr1j[iVar]->FindBin(lny);
	Vr1j[iVar]->SetBinContent(iBin,val);
	Vr1j[iVar]->SetBinError(iBin,err);

	// 2j
	Vr2j[iVar]->Fill(lny,sumN2j[iy][iVar]/sumw[iVar]);
	err = sqrt(sumN2j[iy][iVar])/sumw[iVar];
	iBin = Vr2j[iVar]->FindBin(lny);
	Vr2j[iVar]->SetBinError(iBin,err);
	// 3j
	Vr3j[iVar]->Fill(lny,sumN3j[iy][iVar]/sumw[iVar]);
	err = sqrt(sumN3j[iy][iVar])/sumw[iVar];
	iBin = Vr3j[iVar]->FindBin(lny);
	Vr3j[iVar]->SetBinError(iBin,err);
	// 4j
	Vr4j[iVar]->Fill(lny,sumN4j[iy][iVar]/sumw[iVar]);
	err = sqrt(sumN4j[iy][iVar])/sumw[iVar];
	iBin = Vr4j[iVar]->FindBin(lny);
	Vr4j[iVar]->SetBinError(iBin,err);
	// 5j
	Vr5j[iVar]->Fill(lny,sumN5j[iy][iVar]/sumw[iVar]);
	err = sqrt(sumN5j[iy][iVar])/sumw[iVar];
	iBin = Vr5j[iVar]->FindBin(lny);
	Vr5j[iVar]->SetBinError(iBin,err);
	// 6j
	Vr6j[iVar]->Fill(lny,sumN6j[iy][iVar]/sumw[iVar]);
	err = sqrt(sumN6j[iy][iVar])/sumw[iVar];
	iBin = Vr6j[iVar]->FindBin(lny);
	Vr6j[iVar]->SetBinError(iBin,err);
      }
    }

    // Fill charged-particle histogram
    HnCharged->Fill(1.0*nc,evwt);


    // Periodically update displays
    if (iEvent == 99 || iEvent%500 == 99|| iEvent == nEvent-1) 
      vroot.updateDisplay();       
    if (iEvent == 100 || iEvent%5000==0 || iEvent == nEvent-1) {
      string mcStr="vincia";
      if (not vinciaOn) mcStr="pythia";
      else {
	if (antennaFile.find("Max") != string::npos) mcStr +="-max";
	else if (antennaFile.find("Min") != string::npos) mcStr +="-min";
	else if (antennaFile.find("Ari") != string::npos) mcStr +="-ari";
	else mcStr += "-def";
	if (showerNLC > 0) mcStr += "NLC";
	if (not matching) {
	  if (orderingMode == 3) mcStr+="-m0";
	  else if (orderingMode == 1) mcStr+="-ord";
	}
	else {
	  mcStr += "-m"+num2str(maxMatchTree,1);
	  if (matchingRegOrder != 3) mcStr += "r"+num2str(maxMatchTree,1);
	}
	if (not scaleMatching) mcStr += "-mu0";
	if ( !uncertaintyBands) mcStr+="-nounc";
      }
      if (!settingsPtr->flag("HadronLevel:all")) mcStr+="-nohad";
      
      // Save plots
      vroot.saveDisplay("Rates",mcStr,"eps");
      vroot.saveDisplay("Resolutions",mcStr,"eps");
      vroot.saveDisplay("Angles",mcStr,"eps");

      // Save Pythia histograms for overplotting later
      if (! vinciaOn) {
	vroot.saveHistogram(Vr1j[0],mcStr,"r1j","dat");
	vroot.saveHistogram(Vr2j[0],mcStr,"r2j","dat");
	vroot.saveHistogram(Vr3j[0],mcStr,"r3j","dat");
	vroot.saveHistogram(Vr4j[0],mcStr,"r4j","dat");
	vroot.saveHistogram(Vr5j[0],mcStr,"r5j","dat");
	vroot.saveHistogram(Vr6j[0],mcStr,"r6j","dat");
	vroot.saveHistogram(VLnY12[0],mcStr,"lny12","dat");
	vroot.saveHistogram(VLnY23[0],mcStr,"lny23","dat");
	vroot.saveHistogram(VLnY34[0],mcStr,"lny34","dat");
	vroot.saveHistogram(VLnY45[0],mcStr,"lny45","dat");
	vroot.saveHistogram(VLnY56[0],mcStr,"lny56","dat");
	vroot.saveHistogram(VChiBZ[0],mcStr,"chiBZ","dat");
	vroot.saveHistogram(VThetaNR[0],mcStr,"thetaNR","dat");
	vroot.saveHistogram(VAlpha34[0],mcStr,"alpha34","dat");
	vroot.saveHistogram(VPhiKSW[0],mcStr,"phiKSW","dat");
      }

    }

    // Now only keep exclusive 4-jet events
    if (sorted.size() != 4) continue;

    // For the first event, print the FastJet details
    if (firstEvent) {
      cout << "Ran " << jetDef->description() << endl;
      cout << "Strategy adopted by FastJet was "
           << clustSeq.strategy_string() << endl << endl;      
      firstEvent = false;
    }

    // Compute 4-jet angles
    Vec4 p1(sorted[0].px(),sorted[0].py(),sorted[0].pz(),sorted[0].e());
    Vec4 p2(sorted[1].px(),sorted[1].py(),sorted[1].pz(),sorted[1].e());
    Vec4 p3(sorted[2].px(),sorted[2].py(),sorted[2].pz(),sorted[2].e());
    Vec4 p4(sorted[3].px(),sorted[3].py(),sorted[3].pz(),sorted[3].e());
    // Bengtsson-Zerwas
    Vec4 v1 = cross3(p1,p2);
    Vec4 v2 = cross3(p3,p4);
    double cosChiBZ = abs(costheta(v1,v2));
    // KSW
    v1 = cross3(p1,p4);
    v2 = cross3(p2,p3);
    double alpha1 = theta(v1,v2);
    v1 = cross3(p1,p3);
    v2 = cross3(p2,p4);
    double alpha2 = theta(v1,v2);
    double cosPhiKSW = cos((alpha1+alpha2)/2.);
    // Nachtman-Reiter
    v1 = p1 - p2;
    v2 = p3 - p4;
    double cosThetaNR = abs(costheta(v1,v2));
    // Alpha34
    double cosA34 = costheta(p3,p4);

    // Fill 4-jet angle histograms
    for (int iVar=0; iVar < nVariations; iVar++) {
      VChiBZ[iVar]->Fill(cosChiBZ,vincia.weight(iVar)*wBZ);
      VThetaNR[iVar]->Fill(cosThetaNR,vincia.weight(iVar)*wNR);
      VAlpha34[iVar]->Fill(cosA34,vincia.weight(iVar)*wA34);
      VPhiKSW[iVar]->Fill(cosPhiKSW,vincia.weight(iVar)*wKSW);
    }
  }

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 2) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.statistics();
  if (vinciaOn) vincia.printInfo();

  // Close ROOT file
  cout<<"Saving root file... "<<outfile.str()<<endl;
  // Open ROOT file
  TFile f(outfile.str().c_str(),"recreate");
  // Opened root file:
  cout <<"vincia04-root : opened ROOT file "<<outfile.str()<<endl;  
  VLnY34[0]->Write();
  VLnY45[0]->Write();
  f.Close();

  cout << endl << "Enter q to terminate..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
